import random

URANDOM = open("/dev/urandom", "rb")


def rand_bytes(size: int=16) -> bytes:
    """
    Reads bytes from /dev/urandom.

    Parameters:
        size (int): Number of bytes to read.
    
    Returns:
        bytes: Random bytes.
    """
    return URANDOM.read(size)



def shuffle(in_list: list) -> list:
    """
    Shuffles a list in place using random numbers generated from /dev/urandom.

    Parameters:
        in_list (list): List to be shuffled.
    
    Returns:
        list: in_list shuffled in place.
    """
    random.shuffle(in_list, lambda: int.from_bytes(rand_bytes(32), 'big') / (2 ** 256))



def de_bruijn(k: list, n: int) -> str:
    """
    Generates a de Bruijn sequence for alphabet `k` and subsequences of length `n`.

    Parameters:
        k (list): Alphabet.
        n  (int): Subsequence length.
    
    Returns:
        str: de Bruijn sequence.
    """
    try:
        # let's see if k can be cast to an integer;
        # if so, make our alphabet a list
        _ = int(k)
        alphabet = list(map(str, range(k)))

    except (ValueError, TypeError):
        alphabet = k
        k = len(k)

    a = [0] * k * n
    sequence = []

    def db(t, p):
        if t > n:
            if n % p == 0:
                sequence.extend(a[1:p + 1])
        else:
            a[t] = a[t - p]
            db(t + 1, p)
            for j in range(a[t - p] + 1, k):
                a[t] = j
                db(t + 1, t)
    db(1, 1)
    return "".join(alphabet[i] for i in sequence)



from samson.block_ciphers.modes.cbc import CBC
from samson.utilities.bytes import Bytes
from samson.block_ciphers.rijndael import Rijndael

class CBCMAC(object):
    """
    Message authentication code scheme based off of a block cipher in CBC mode.
    """

    def __init__(self, key: bytes, cipher: Rijndael, iv: bytes=b'\x00' * 16):
        """
        Parameters:
            key    (bytes): Bytes-like object to key the underlying cipher.
            cipher (class): Instantiable class representing a block cipher.
            iv     (bytes): Initialization vector for CBC mode.
        """
        self.key = key
        self.iv = iv
        self.cipher = cipher

    
    def __repr__(self):
        return f"<CBCMAC: key={self.key}, iv={self.iv}, cipher={self.cipher}>"


    def __str__(self):
        return self.__repr__()

    
    def generate(self, message: bytes, pad: bool=True) -> Bytes:
        """
        Generates a keyed MAC for `message`.

        Parameters:
            message (bytes): Message to generate a MAC for.
            pad      (bool): Whether or not to pad the message with PKCS7.
        
        Returns:
            Bytes: The MAC.
        """
        cryptor = self.cipher(self.key)
        cbc = CBC(cryptor.encrypt, cryptor.decrypt, self.iv, cryptor.block_size // 8)
        return cbc.encrypt(message, pad)[-(cryptor.block_size // 8):]


from .default_oracle import DefaultOracle
from .encryption_oracle import EncryptionOracle
from .padding_oracle import PaddingOracle
from .stateless_block_encryption_oracle import StatelessBlockEncryptionOracle


__all__ = ["DefaultOracle", "EncryptionOracle", "PaddingOracle", "StatelessBlockEncryptionOracle"]


from types import FunctionType

class EncryptionOracle(object):
    """
    Oracle that provides an interface to a chosen-plaintext attack.
    """

    def __init__(self, request_func: FunctionType):
        """
        Parameters:
            request_func (func): Function that takes in bytes and returns a ciphertext.
        """
        self.encrypt = request_func


from samson.utilities.general import rand_bytes

class MarkovState(object):
    """
    Represents a state in a Markov chain.
    """
    
    def __init__(self, count: int, probability: float, transitions: dict):
        """
        Parameters:
            count          (int): Number of hits on this state.
            probabilitiy (float): Probability of transitioning to this state.
            transitions   (dict): Dictionary of {str: MarkovState} as possible transitions.
        """
        self.count = count
        self.probability = probability
        self.transitions = transitions


    def __repr__(self):
        return "<MarkovState: transitions={}, count={}, probability={}>".format(self.transitions.keys(), self.count, self.probability)

    def __str__(self):
        return self.__repr__()


    def calculate_chain_probs(self):
        """Recursively turn counts into probabilities"""
        total_count = sum([subchain.count for token, subchain in self.transitions.items()])

        for _token, subchain in self.transitions.items():
            subchain.probability = subchain.count / total_count
            subchain.calculate_chain_probs()



    def random_walk(self, distance: int) -> list:
        """
        Randomly walks `distance` down the chain and returns the list of transition tokens.

        Parameters:
            distance (int): Distance to walk down the chain.

        Returns:
            list: List of transition tokens from the walk.
        """
        result = []
        if distance > 0:
            rand_num = int.from_bytes(rand_bytes(1), 'big')
            winning_probability = rand_num / 256
            
            prob_accumulator = 0.0

            for token, subchain in self.transitions.items():
                prob_accumulator += subchain.probability

                if prob_accumulator >= winning_probability:
                    result = [token] + subchain.random_walk(distance - 1)
                    break

        return result



from samson.prngs.bitsliced_flfsr import BitslicedFLFSR
from samson.utilities.bytes import Bytes

# https://github.com/ttsou/airprobe/blob/master/A5.1/python/A51_Tables/a51.py
# Implemented in big endian
class A51(object):
    """
    A5/1 stream cipher

    Used in GSM celluar encryption.
    """

    def __init__(self, key: bytes, frame_num: int):
        """
        Parameters:
            key     (bytes): Key (64 bits).
            frame_num (int): Current frame number (22 bits).
        """
        self.key = key
        self.frame_num = frame_num

        self.lfsr_regs = [
            BitslicedFLFSR(19, 10, [13, 16, 17, 18]),
            BitslicedFLFSR(22, 11, [20, 21]),
            BitslicedFLFSR(23, 12, [7, 20, 21, 22])
        ]

        for lfsr in self.lfsr_regs:
            lfsr.mix_state(self.key, 64)
            lfsr.mix_state(self.frame_num, 22)

        for _ in range(100):
            self.clock()

       
       
    def __repr__(self):
        return f"<A51: key={self.key}, frame_num={self.frame_num}, lfsr_regs={self.lfsr_regs}>"

    def __str__(self):
        return self.__repr__()



    def clock(self):
        """
        Performs the majority-vote clocking.
        """
        majority = sum([lfsr.clock_value() for lfsr in self.lfsr_regs]) // 2

        for lfsr in self.lfsr_regs:
            if lfsr.clock_value() == majority:
                lfsr.clock()


    
    def generate(self, length: int) -> Bytes:
        """
        Generates `length` of keystream.

        Parameters:
            length (int): Desired length of keystream in bytes.
        
        Returns:
            Bytes: Keystream.
        """
        bitstring = ''
        for _ in range(length * 8):
            self.clock()
            bitstring += str(self.lfsr_regs[0].value() ^ self.lfsr_regs[1].value() ^ self.lfsr_regs[2].value())
        
        return Bytes(int(bitstring, 2))


from .davies_meyer_construction import DaviesMeyerConstruction
from .even_mansour_scheme import EvenMansourScheme
from .feistel_network import FeistelNetwork
from .matyas_meyer_oseas_construction import MatyasMeyerOseasConstruction
from .merkle_damgard_construction import MerkleDamgardConstruction
from .sponge_construction import SpongeConstruction


__all__ = ["DaviesMeyerConstruction", "EvenMansourScheme", "FeistelNetwork", "MatyasMeyerOseasConstruction", "MerkleDamgardConstruction", "SpongeConstruction"]


import struct
from samson.utilities.manipulation import left_rotate
from samson.constructions.merkle_damgard_construction import MerkleDamgardConstruction
from samson.hashes.md5 import state_to_bytes, bytes_to_state
from samson.utilities.bytes import Bytes


def F(x,y,z):
    return (x & y) | (~x & z)

def G(x,y,z):
    return (x & y) | (x & z) | (y & z)

def H(x,y,z):
    return x ^ y ^ z

iv = [
        0x67452301,
        0xefcdab89,
        0x98badcfe,
        0x10325476
    ]



def compression_func(message, state):
    X = list( struct.unpack("<16I", message) + (None,) * (80-16) )
    #h = [x for x in bytes_to_state(state)]
    h = bytes_to_state(state)
    last_state = [x for x in h]

    # Round 1
    s = (3,7,11,19)
    for r in range(16):
        i = (16-r)%4
        k = r
        h[i] = left_rotate( (h[i] + F(h[(i+1)%4], h[(i+2)%4], h[(i+3)%4]) + X[k]) % 2**32, s[r%4] )


    # Round 2
    s = (3,5,9,13)
    for r in range(16):
        i = (16-r)%4
        k = 4*(r%4) + r//4
        h[i] = left_rotate( (h[i] + G(h[(i+1)%4], h[(i+2)%4], h[(i+3)%4]) + X[k] + 0x5a827999) % 2**32, s[r%4] )


    # Round 3
    s = (3,9,11,15)
    k = (0,8,4,12,2,10,6,14,1,9,5,13,3,11,7,15)
    for r in range(16):
        i = (16-r)%4
        h[i] = left_rotate( (h[i] + H(h[(i+1)%4], h[(i+2)%4], h[(i+3)%4]) + X[k[r]] + 0x6ed9eba1) % 2**32, s[r%4] )

    new_state = []
    for i,v in enumerate(h):
        new_state.append((v + last_state[i]) % 2**32)
    
    return Bytes(state_to_bytes(new_state))



class MD4(MerkleDamgardConstruction):
    """
    Obsolete cryptographic hash function and predecessor to MD5.
    """

    def __init__(self, initial_state: bytes=state_to_bytes(iv)):
        """
        Parameters:
            initial_state (bytes): (Optional) Initial internal state.
        """
        super().__init__(
            initial_state=initial_state,
            compression_func=compression_func,
            digest_size=16,
            endianness='little'
        )


    def __repr__(self):
        return f"<MD4: initial_state={self.initial_state}, block_size={self.block_size}>"


    def __str__(self):
        return self.__repr__()


from samson.utilities.bytes import Bytes

class DGHVBit(int):
    """
    Convenience class for DGHV bit operations.
    """

    def __new__(cls, val, *args, **kwargs):
        return super(DGHVBit, cls).__new__(cls, val)

    def NOT(self):
        return DGHVBit(1 + self)


    def AND(self, b: int):
        return DGHVBit(self * b)


    def NAND(self, b: int):
        return DGHVBit(self.AND(b).NOT())


    def OR(self, b: int):
        return self.NAND(self).NAND(b.NAND(b))


    def XOR(self, b: int):
        return DGHVBit(self.NOT() * b + b.NOT() * self)


    def __invert__(self):
        return self.NOT()

    
    def __or__(self, b: int):
        return self.OR(b)


    def __and__(self, b: int):
        return self.AND(b)

    
    def __xor__(self, b: int):
        return self.XOR(b)
    


class DGHV(object):
    """
    The Dijk-Gentry-Halevi-Vaikuntanathan (DGHV) fully-homomorphic encryption scheme
    """
    
    def __init__(self, p: int=None):
        """
        Parameters:
            p (int): Key.
        """
        self.p = p or Bytes.random(8).to_int()


    def __repr__(self):
        return f"<DGHV: p={self.p}>"

    def __str__(self):
        return self.__repr__()



    def encrypt(self, m: int) -> DGHVBit:
        """
        Encrypts a bit `m`.

        Parameters:
            m (int): Number to encrypt.

        Returns:
            DGHVBit: Encrypted bit.
        """
        q = Bytes.random(8).to_int()
        r = Bytes.random(8).to_int() % (self.p // 4)
        return DGHVBit(self.p * q + 2 * r + m)


    
    def decrypt(self, c: int) -> int:
        """
        Decrypts a number `c` into a plaintext bit.

        Parameters:
            c (int): Encrypted int or DGHVBit.
        
        Returns:
            int: Plaintext bit.
        """
        return (c % self.p) % 2



from samson.utilities.analysis import generate_rc4_bias_map, RC4_BIAS_MAP
from samson.oracles.encryption_oracle import EncryptionOracle
from samson.utilities.bytes import Bytes
import multiprocessing
import itertools
import struct

import logging
log = logging.getLogger(__name__)



class RC4PrependAttack(object):
    """
    Performs a plaintext recovery attack.

    The RC4PrependAttack uses an EncryptionOracle that prepends the payload and then encrypts the whole message with a randomly generated key.
    The attack uses specific byte biases found in the RC4 keystream and feeds the plaintext into these locations. Using the law of large numbers,
    the keystream bias will surface, and we can XOR it with the ciphertext to return the plaintext. This specific implementation uses several optmizations.
    The first is concurrency/parallelism and is self-explanatory. The second is the use of concurrent biases. We first determine which biases we can still use
    in the attack. This is necessary since we can only feed the plaintext forward, so plaintext index five is not eligible to use RC4 bias index two. Next, we determine
    which biases are "active". If the plaintext is 20 characters long, and the current padding position is 0, then we can utilize both RC4 bias indices 2 and 16 (i.e. 1 and 15 using a zero-based-index).
    Lastly, we introduce "branches". If a byte happens to decrypted more than one and has different outcomes, we keep both. At the end, we return the Cartesian product of all
    byte decryptions. This will normally just be one.

    Conditions:
        * RC4 is being used
        * The user has access to an oracle that encrypts user-controlled plaintext and a secret under a random key
        * The user-controlled plaintext is prepended to the secret
    """

    def __init__(self, oracle: EncryptionOracle):
        """
        Parameters:
            oracle (EncryptionOracle): An oracle that takes in arbitrary plaintext bytes and returns its encryption under a random key.
        """
        self.oracle = oracle
        self.strongest_biases = [1, 15, 31]
        

    def _encrypt_chunk(self, payload: bytes, chunk_size: int):
        return [self.oracle.encrypt(payload) for _ in range(chunk_size)]


    def execute(self, secret_length: int, sample_size: int=2**23) -> Bytes:
        """
        Executes the attack.

        Parameters:
            secret_length (int): The length of the secret you're trying to recover.
            sample_size   (int): The amount of samples to collect per byte of the secret. Higher numbers are slower but more accurate.
        
        Returns:
            Bytes: The recovered plaintext.
        """
        cracked_indices = [set() for i in range(secret_length)]
        cpu_count = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=cpu_count)

        log.debug("Running with {} cores".format(cpu_count))

        for i in range(secret_length):
            log.debug("Starting iteration {}/{}".format(i + 1, secret_length))

            if len(cracked_indices[i]) > 0:
                continue

            applicable_biases = [bias for bias in self.strongest_biases if i <= bias or (secret_length + i >= bias and i < bias)]
            padding_len = max(applicable_biases[0] - i, 0)
            active_biases = [bias for bias in applicable_biases if padding_len + secret_length > bias]

            payload = b'\x00' * padding_len
            chunk_size = sample_size // cpu_count

            log.debug("Sampling {} ciphertexts".format(sample_size))
            random_ciphertexts = [pool.apply_async(self._encrypt_chunk, (payload, chunk_size)) for i in range(cpu_count)]
            flattened_list = [result for result_list in random_ciphertexts for result in result_list.get()]

            log.debug("Generating bias map")
            bias_map = generate_rc4_bias_map(flattened_list)

            for bias_idx in active_biases:
                cracked_indices[bias_idx - padding_len].add(RC4_BIAS_MAP[bias_idx] ^ bias_map[bias_idx][0][0])


        all_branches = itertools.product(*[list(results) for results in cracked_indices])
        return [Bytes(struct.pack('B' * len(branch), *branch)) for branch in all_branches]


#!/usr/bin/python3
w, n, m, r = (32, 624, 397, 31)
a = 0x9908b0df
f = 1812433253
u, d = (11, 0xFFFFFFFF)
s, b = (7, 0x9D2C5680)
t, c = (15, 0xEFC60000)
l = 18


def asint32(integer):
    return integer & d

def temper(y):
    y ^= (y >> u)
    y ^= (y << s) & b
    y ^= (y << t) & c
    y ^= (y >> l)
    return y


# We don't have to include a constant since 32-bit Mersenne Twister doesn't
# use non-idempotent constants on its right shifts.
def _untemper_right(y, bits):
    # Create a 32-bit mask with `bits` 1s at the beginning.
    # We'll shift this over the iterations to invert the temper.
    mask = (1 << bits) - 1 << 32 - bits
    shift_mod = 0

    while mask > 0:
        y ^= shift_mod

        # Get next `bits` bits of y
        # Ex: bits = 3, mask = '00011100000000000'
        # '100_010_00100100001' -> '000000_010_00000000'
        shift_mod = (y & mask) >> bits

        # Move mask right `bits` to select next bits to shift
        # Ex: bits = 3
        # 11100000000000000 -> 00011100000000000
        mask >>= bits
    return y


def _untemper_left(y, bits, constant):
    int32_mask = 0xFFFFFFFF
    mask = (1 << bits) - 1
    shift_mod = 0

    while (mask & int32_mask) > 0:
        y ^= shift_mod & constant

        # Get next `bits` bits of y
        shift_mod = (y & mask) << bits

        # Move mask right `bits` to select next bits to shift
        mask <<= bits
    return y


def _untemper(y):
    y = _untemper_right(y, l)
    y = _untemper_left(y, t, c)
    y = _untemper_left(y, s, b)
    y = _untemper_right(y, u)
    return y & d



# Implementation of MT19937
class MT19937:
    """
    Mersenne Twister 19937
    """

    def __init__(self, seed: int=0):
        """
        Parameters:
            seed (int): Initial value.
        """
        self.state = [0] * n
        self.seed = seed

        # Seed the algo
        self.index = n
        self.state[0] = seed

        for i in range(1, n):
            self.state[i] = asint32(f * (self.state[i - 1] ^ self.state[i - 1] >> (w - 2)) + i)


    def __repr__(self):
        return f"<MT19937: seed={self.seed}, index={self.index}, state={self.state}>"


    def __str__(self):
        return self.__repr__()
        

    def twist(self):
        """
        Called internally. Performs the `twist` operation of the Mersenne Twister.
        """
        for i in range(n):
            y = asint32((self.state[i] & 0x80000000) + (self.state[(i + 1) % n] & 0x7fffffff))
            self.state[i] = self.state[(i + m) % n] ^ y >> 1

            if y & 1 == 1:
                self.state[i] ^= a

        self.index = 0


    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        if self.index >= n:
            self.twist()

        y = self.state[self.index]
        y  = temper(y)

        self.index += 1

        return asint32(y)


    @staticmethod
    def crack(observed_outputs: list):
        """
        Given 624 observed, consecutive outputs, cracks the internal state of the original and returns a replica.

        Parameters:
            observed_outputs (list): List of consecutive inputs (in order, obviously).

        Returns:
            MT19937: A replica of the original MT19937.
        """
        if len(observed_outputs) < 624:
            raise ValueError("`observed_outputs` must contain at least 624 consecutive outputs.")

        cloned = MT19937(0)
        cloned.state = [_untemper(output) for output in observed_outputs][-624:]

        return cloned


from samson.utilities.manipulation import left_rotate
from types import FunctionType

MASK58 = 0x3FFFFFFFFFFFFFF
MASK64 = 0xFFFFFFFFFFFFFFFF

def V116_PLUS(x):
    s0, s1 = x
    result = (s0 + s1) & MASK58

    s1 ^= s0
    x[0] = left_rotate(s0, 24, bits=58) ^ s1 ^ ((s1 << 2) & MASK58)
    x[1] = left_rotate(s1, 35, bits=58)
    return x, result


# Might be correct for V128.
# def V128(x):
#     s0, s1 = x
#     result = (s0 + s1) & MASK64

#     s1 ^= s0
#     x[0] = left_rotate(s0, 55, bits=64) ^ s1 ^ ((s1 << 14) & MASK64)
#     x[1] = left_rotate(s1, 36, bits=64)
#     return x, result


def V128_PLUS(x):
    s0, s1 = x
    result = (s0 + s1) & MASK64

    s1 ^= s0
    x[0] = left_rotate(s0, 24, bits=64) ^ s1 ^ ((s1 << 16) & MASK64)
    x[1] = left_rotate(s1, 37, bits=64)
    return x, result


class Xoroshiro(object):
    def __init__(self, seed: int, variant: FunctionType=V128_PLUS):
        """
        Parameters:
            seed       (int): Initial value.
            variant   (func): Xoroshiro variant. Function that takes in an integer and returns (new_state, output).
        """
        self.state = seed
        self.variant = variant

    def __repr__(self):
        return f"<Xoroshiro: state={self.state}, variant={self.variant}>"


    def __str__(self):
        return self.__repr__()
    

    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.state, result = self.variant(self.state)
        return result


from samson.constructions.merkle_damgard_construction import MerkleDamgardConstruction
from samson.utilities.bytes import Bytes
from samson.utilities.manipulation import right_rotate, get_blocks

# https://en.wikipedia.org/wiki/SHA-2
H_256 = [0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19]
H_224 = [0xc1059ed8, 0x367cd507, 0x3070dd17, 0xf70e5939, 0xffc00b31, 0x68581511, 0x64f98fa7, 0xbefa4fa4]
H_512 = [0x6a09e667f3bcc908, 0xbb67ae8584caa73b, 0x3c6ef372fe94f82b, 0xa54ff53a5f1d36f1, 0x510e527fade682d1, 0x9b05688c2b3e6c1f, 0x1f83d9abfb41bd6b, 0x5be0cd19137e2179]
H_384 = [0xcbbb9d5dc1059ed8, 0x629a292a367cd507, 0x9159015a3070dd17, 0x152fecd8f70e5939, 0x67332667ffc00b31, 0x8eb44a8768581511, 0xdb0c2e0d64f98fa7, 0x47b5481dbefa4fa4]

ROT_256 = [7, 18, 3, 17, 19, 10, 6, 11, 25, 2, 13, 22]
ROT_512 = [1, 8, 7, 19, 61, 6, 14, 18, 41, 28, 34, 39]

K_256 = [
   0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
   0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
   0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
   0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
   0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
   0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
   0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
   0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]

K_512 = [
    0x428a2f98d728ae22, 0x7137449123ef65cd, 0xb5c0fbcfec4d3b2f, 0xe9b5dba58189dbbc, 0x3956c25bf348b538, 
    0x59f111f1b605d019, 0x923f82a4af194f9b, 0xab1c5ed5da6d8118, 0xd807aa98a3030242, 0x12835b0145706fbe, 
    0x243185be4ee4b28c, 0x550c7dc3d5ffb4e2, 0x72be5d74f27b896f, 0x80deb1fe3b1696b1, 0x9bdc06a725c71235, 
    0xc19bf174cf692694, 0xe49b69c19ef14ad2, 0xefbe4786384f25e3, 0x0fc19dc68b8cd5b5, 0x240ca1cc77ac9c65, 
    0x2de92c6f592b0275, 0x4a7484aa6ea6e483, 0x5cb0a9dcbd41fbd4, 0x76f988da831153b5, 0x983e5152ee66dfab, 
    0xa831c66d2db43210, 0xb00327c898fb213f, 0xbf597fc7beef0ee4, 0xc6e00bf33da88fc2, 0xd5a79147930aa725, 
    0x06ca6351e003826f, 0x142929670a0e6e70, 0x27b70a8546d22ffc, 0x2e1b21385c26c926, 0x4d2c6dfc5ac42aed, 
    0x53380d139d95b3df, 0x650a73548baf63de, 0x766a0abb3c77b2a8, 0x81c2c92e47edaee6, 0x92722c851482353b, 
    0xa2bfe8a14cf10364, 0xa81a664bbc423001, 0xc24b8b70d0f89791, 0xc76c51a30654be30, 0xd192e819d6ef5218, 
    0xd69906245565a910, 0xf40e35855771202a, 0x106aa07032bbd1b8, 0x19a4c116b8d2d0c8, 0x1e376c085141ab53, 
    0x2748774cdf8eeb99, 0x34b0bcb5e19b48a8, 0x391c0cb3c5c95a63, 0x4ed8aa4ae3418acb, 0x5b9cca4f7763e373, 
    0x682e6ff3d6b2b8a3, 0x748f82ee5defb2fc, 0x78a5636f43172f60, 0x84c87814a1f0ab72, 0x8cc702081a6439ec, 
    0x90befffa23631e28, 0xa4506cebde82bde9, 0xbef9a3f7b2c67915, 0xc67178f2e372532b, 0xca273eceea26619c, 
    0xd186b8c721c0c207, 0xeada7dd6cde0eb1e, 0xf57d4f7fee6ed178, 0x06f067aa72176fba, 0x0a637dc5a2c898a6, 
    0x113f9804bef90dae, 0x1b710b35131c471b, 0x28db77f523047d84, 0x32caab7b40c72493, 0x3c9ebe0a15c9bebc, 
    0x431d67c49c100d4c, 0x4cc5d4becb3e42b6, 0x597f299cfc657e2a, 0x5fcb6fab3ad6faec, 0x6c44198c4a475817
]


class SHA2(MerkleDamgardConstruction):
    """
    SHA2 hash function base class.
    """

    def __init__(self, initial_state: list, digest_size: int, state_size: int, block_size: int, rounds: int, rot: list, k: list):
        """
        Parameters:
            initial_state (list): Initial state as list of integers.
            digest_size    (int): Output size in bytes.
            state_size     (int): Number of elements in state.
            block_size     (int): Amount of message to digest at a time.
            rounds         (int): Number of compression rounds to perform.
            rot           (list): Rotation constants.
            k             (list): `k` constants.
        """
        super().__init__(
            initial_state=Bytes(b''.join([int.to_bytes(h_i, state_size, 'big') for h_i in initial_state])),
            compression_func=None,
            digest_size=digest_size,
            block_size=block_size
        )

        self.state_size = state_size
        self.rounds = rounds
        self.rot = rot
        self.k = k



    def yield_state(self, message: bytes):
        """
        Yields successive states while processing `message`.

        Parameters:
            message (bytes): Message to hash.
        
        Returns:
            generator: Generator yielding states.
        """
        for state in MerkleDamgardConstruction.yield_state(self, message):
            yield state[:self.digest_size]
    


    def __repr__(self):
        return "<SHA2: initial_state={}, block_size={}, digest_size={}>".format(self.initial_state, self.block_size, self.digest_size)

    def __str__(self):
        return self.__repr__()


    def compression_func(self, block: bytes, state: bytes) -> Bytes:
        """
        SHA-2 compression function.

        Parameters:
            block (bytes): Block being digested.
            state (bytes): Current digest state.
        
        Returns:
            Bytes: Hash output.
        """
        bit_mask = 0xFFFFFFFF if self.state_size == 4 else 0xFFFFFFFFFFFFFFFF
        bit_size = self.state_size * 8

        state = [int.from_bytes(chunk, 'big') for chunk in state.chunk(self.state_size)]
        w = [int.from_bytes(b, 'big') for b in get_blocks(block, self.state_size)] + ([None] * (self.rounds - 16))

        for i in range(16, self.rounds):
            s0 = right_rotate(w[i-15], self.rot[0], bit_size) ^ right_rotate(w[i-15], self.rot[1], bit_size) ^ (w[i-15] >> self.rot[2])
            s1 = right_rotate(w[i-2], self.rot[3], bit_size) ^ right_rotate(w[i-2], self.rot[4], bit_size) ^ (w[i-2] >> self.rot[5])
            w[i] = (w[i-16] + s0 + w[i-7] + s1) & bit_mask
        

        a = state[0]
        b = state[1]
        c = state[2]
        d = state[3]
        e = state[4]
        f = state[5]
        g = state[6]
        h = state[7]

        for i in range(self.rounds):
            S1 = right_rotate(e,  self.rot[6], bit_size) ^ right_rotate(e, self.rot[7], bit_size) ^ right_rotate(e, self.rot[8], bit_size)
            ch = g ^ (e & (f ^ g))
            temp1 = (h + S1 + ch + self.k[i] + w[i])
            S0 = right_rotate(a, self.rot[9], bit_size) ^ right_rotate(a, self.rot[10], bit_size) ^ right_rotate(a, self.rot[11], bit_size)
            maj = (a & b) ^ (a & c) ^ (b & c)
            temp2 = (S0 + maj)

            h = g
            g = f
            f = e
            e = (d + temp1) & bit_mask
            d = c
            c = b
            b = a
            a = (temp1 + temp2) & bit_mask


        state[0] += a
        state[1] += b
        state[2] += c
        state[3] += d
        state[4] += e
        state[5] += f
        state[6] += g
        state[7] += h

        return Bytes(b''.join([int.to_bytes(h_i & bit_mask, self.state_size, 'big') for h_i in state]))

    
class SHA224(SHA2):
    def __init__(self, h: list=None):
        """
        Parameters:
            h (list): Initial state as list of integers.
        """
        super().__init__(
            initial_state=h or H_224,
            digest_size=224 // 8,
            state_size=4,
            block_size=64,
            rounds=64,
            rot=ROT_256,
            k=K_256
        )


class SHA256(SHA2):
    def __init__(self, h: list=None):
        """
        Parameters:
            h (list): Initial state as list of integers.
        """
        super().__init__(
            initial_state=h or H_256,
            digest_size=256 // 8,
            state_size=4,
            block_size=64,
            rounds=64,
            rot=ROT_256,
            k=K_256
        )


class SHA384(SHA2):
    def __init__(self, h: list=None):
        """
        Parameters:
            h (list): Initial state as list of integers.
        """
        super().__init__(
            initial_state=h or H_384,
            digest_size=384 // 8,
            state_size=8,
            block_size=128,
            rounds=80,
            rot=ROT_512,
            k=K_512
        )


class SHA512(SHA2):
    def __init__(self, h: list=None):
        """
        Parameters:
            h (list): Initial state as list of integers.
        """
        super().__init__(
            initial_state=h or H_512,
            digest_size=512 // 8,
            state_size=8,
            block_size=128,
            rounds=80,
            rot=ROT_512,
            k=K_512
        )



import string
import struct

import logging
log = logging.getLogger(__name__)

alphabet = (string.ascii_lowercase + string.ascii_uppercase + string.digits + '=/+' + ':;').encode()
padding_chars = ')(\{\}\\[]<%@#^`~$*!|`'[::-1].encode() + b'>.\'_,?;:+-='


class CompressionRatioSideChannelAttack(object):
    def __init__(self, oracle, alphabet=alphabet, padding_chars=padding_chars, block_size=16):
        self.oracle = oracle
        self.block_size = block_size
        self.alphabet = alphabet
        self.padding_chars = padding_chars


    def execute(self, known_plaintext, secret_len):
        plaintext = known_plaintext
        extra_spaces = b' ' * self.block_size * 4

        for pos in range(secret_len - len(known_plaintext)):
            request_lengths = {}
            # padding = self._find_padding((plaintext + b'~' + extra_spaces) * (self.block_size // 2))
            padding = self._find_padding((plaintext + b'z' + b'z' * self.block_size * 4) * (self.block_size // 2))

            for curr_byte in self.alphabet:
                as_byte = struct.pack('B', curr_byte)
                payload = plaintext + as_byte

                #print(padding)
                # new_length = self.oracle.request(padding + (payload + extra_spaces) * (self.block_size // 2))
                new_length = self.oracle.request(padding + (payload + b'z' * self.block_size * 4) * (self.block_size // 2))
                request_lengths[as_byte] = new_length

            sorted_lengths = sorted(request_lengths.items(), key=lambda item: item[1])
            plaintext += sorted_lengths[0][0]

            if sorted_lengths[0][1] == sorted_lengths[1][1]:
                log.warning('Multiple best answers found')
        return plaintext


    def _find_padding(self, payload):
        # Unlikely characters. NOTE: Since we only include each character once
        # (to prevent compression), we need a suitable number of characters to actually
        # "pad" the CBC block to force it over a byte boundary. Sixteen should always work
        # for 16-byte blocks (e.g. AES-128).
        #possible_padding = ')(\{\}\\[]<%@#^`~$*!|`'[::-1].encode() + b'>.\'_,?;:+-='
        padding = b''

        log.debug('Attempting to find suitable padding')

        original_length = self.oracle.request(payload)

        for char in self.padding_chars:
            padding += struct.pack('B', char)

            if self.oracle.request(padding + payload) > original_length:
                log.debug('Found padding: {}'.format(padding.decode()))
                return padding
        raise Exception("Couldn't find a suitable padding")


from samson.utilities.bytes import Bytes
import math

# https://tools.ietf.org/html/rfc8017#section-7.2.1
class PKCS1v15(object):
    """
    PCKS#1 v1.5 RSA padding
    """

    def __init__(self, key_bit_length: int):
        """
        Parameters:
            key_bit_length (int): Length of the RSA modulus in bits.
        """
        self.key_byte_length = math.ceil(key_bit_length / 8)
    

    def __repr__(self):
        return f"<PKCS1v15: key_byte_length={self.key_byte_length}>"

    def __str__(self):
        return self.__repr__()


    def pad(self, plaintext: bytes) -> Bytes:
        """
        Pads the plaintext.

        Parameters:
            plaintext (bytes): Plaintext to pad.
        
        Returns:
            Bytes: Padded plaintext.
        """
        pad_len = self.key_byte_length - 3 - len(plaintext)
        assert pad_len >= 8

        padding = Bytes.random(pad_len) | Bytes(0x01).stretch(pad_len)
        return b'\x00\x02' + padding + b'\x00' + plaintext



    def unpad(self, plaintext: bytes, allow_padding_oracle: bool=False) -> Bytes:
        """
        Unpads the plaintext.

        Parameters:
            plaintext           (bytes): Plaintext to unpad.
            allow_padding_oracle (bool): Whether or not to explicitly create a padding oracle.
        
        Returns:
            Bytes: Unpadded plaintext.
        """
        if allow_padding_oracle and (plaintext[:2] != b'\x00\x02' or len(plaintext) != self.key_byte_length):
            raise Exception('Invalid padding ;)')

        header_removed = plaintext[2:]
        return header_removed[header_removed.index(b'\x00') + 1:]


from samson.utilities.math import modexp
from samson.protocols.srp_client import SRPClient
from samson.hashes.sha2 import SHA256
from samson.utilities.bytes import Bytes

class SRPServer(object):
    """
    Secure Remote Password protocol server
    """

    def __init__(self, g: int=2, N: int=SRPClient.MODP_1024, hash_obj: object=SHA256(), b: int=None):
        """
        Parameters:
            g           (int): Generator.
            N           (int): Prime modulus.
            hash_obj (object): Instantiated object with compatible hash interface
            b           (int): Random private value.
        """
        self.g = g
        self.N = N
        self.hash_obj = hash_obj

        self.accounts = {}
        self.requests = {}
        self.salt = Bytes.random(4)

        self.b = b or Bytes.random(4).int() % N
        self.k = hash_obj.hash(Bytes(N) + self.PAD(g)).int()


    def PAD(self, in_bytes: bytes) -> Bytes:
        """
        If a conversion is explicitly specified with the operator PAD(), the integer will first be implicitly converted, then the resultant byte-string will be left-padded with zeros (if necessary) until its length equals the implicitly-converted length of N.
        """
        return Bytes.wrap(in_bytes).zfill((self.N.bit_length() + 7) // 8)




    def create_account(self, identity: bytes, password: bytes):
        """
        Creates a new account entry with the server.

        Parameters:
            identity (bytes): Username.
            password (bytes): Password.
        """
        # x = self.hash_obj.hash(self.salt + password).int()
        x = self.hash_obj.hash(self.salt + self.hash_obj.hash(identity + b':' + password)).int()
        v = modexp(self.g, x, self.N)
        self.accounts[identity] = v


    
    def respond_with_challenge(self, identity: bytes, A: int) -> (bytes, int):
        """
        Receives the client's challenge and returns the server's challenge.

        Parameters:
            identity (bytes): Username.
            A          (int): Client's challenge.

        Returns:
            (bytes, int): Formatted as (server salt, server's challenge `B`)
        """
        v = self.accounts[identity]
        B = (self.k * v + modexp(self.g, self.b, self.N)) % self.N
        self.requests[identity] = {'A': A, 'B': B}
        
        return self.salt, B



    def check_challenge(self, identity: bytes, client_hash: bytes) -> bool:
        """
        Checks if the client's challenge is correct.

        Parameters:
            identity    (bytes): Username.
            client_hash (bytes): Client's hash challenge.
        
        Returns:
            bool: Whether or not the challenge is correct.
        """
        request = self.requests[identity]
        v = self.accounts[identity]
        A = request['A']

        uH = self.hash_obj.hash(self.PAD(A) + self.PAD(request['B'])).int()

        p1 = (A * modexp(v, uH, self.N))
        sS = modexp(p1, self.b, self.N)

        sK = self.hash_obj.hash(Bytes(sS))
        return self.hash_obj.hash(sK + self.salt) == client_hash


# https://stackoverflow.com/questions/14519227/rail-fence-cipher-looking-for-a-better-solution
class RailFence(object):
    """
    Zigzag transposition cipher whose construction resembles a rail fence.
    """

    def __init__(self, num_rails: int):
        """
        Parameters:
            num_rails (int): Number of "fence rails" to encrypt/decrypt over. This is effectively the key.
        """
        self.num_rails = num_rails



    def _fence(self, in_list):
        fence = [[None] * len(in_list) for n in range(self.num_rails)]
        rails = list(range(self.num_rails - 1)) + list(range(self.num_rails - 1, 0, -1))

        for n, x in enumerate(in_list):
            fence[rails[n % len(rails)]][n] = x

        return [c for rail in fence for c in rail if c is not None]



    def encrypt(self, plaintext: str) -> str:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (str): String to be encrypted.
        
        Returns:
            str: Resulting ciphertext.
        """
        return ''.join(self._fence(plaintext))


    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (str): Stringt to be decrypted.
        
        Returns:
            str: Resulting plaintext.
        """
        enc_pos = self._fence(range(len(ciphertext)))
        return ''.join([ciphertext[enc_pos.index(i)] for i in range(len(ciphertext))])
        


from samson.utilities.manipulation import xor_buffs
from samson.analyzers.analyzer import Analyzer

import logging
log = logging.getLogger(__name__)

# TODO: Make work with more than two ciphertexts.
class XORDictionaryAttack(object):
    """
    Preforms a plaintext recovery attack.

    Attempts to retrieve the plaintext from two or more ciphertexts using a dictionary attack.
    The basic premise is if the ciphertexts are fully contained within the wordlist,
    then XORing one ciphertext with the correct word (or whatever values are in the wordlist)
    will produce the correct keystream segment. This segment is then XOR'd with the other ciphertext and
    fed through the analyzer. The word with the highest score is most likely correct.

    Conditions:
        * A stream/OTP-like cipher is used. I.E. plaintext XOR keystream
        * The user has collected more than one ciphertext using the same keystream.
    """

    def __init__(self, analyzer: Analyzer, wordlist: list):
        """
        Parameters:
            analyzer (Analyzer): Analyzer that correctly scores the underlying plaintext.
            wordlist     (list): List of strings for the dictionary attack.
        """
        self.analyzer = analyzer
        self.wordlist = wordlist
        
    
    def execute(self, ciphertexts: list, word_ranges: list=[2,3,4], delimiter: str=' ') -> list:
        """
        Executes the attack.
        
        Parameters:
            ciphertexts (list): List of bytes-like ciphertexts using the same keystream.
            word_ranges (list): List of numbers of words to try. E.G. [2, 3, 4] means
                                try the Cartesian product of 2, 3,
                                and 4-tuple word combinations.
            delimiter    (str): Delimiter to use between word combinations.
        
        Returns:
            list: Top 10 possible plaintexts.
        """

        if len(ciphertexts) != 2:
            raise ValueError('`ciphertexts` MUST contain two samples.')

        two_time = xor_buffs(*ciphertexts)

        cipher_len = len(two_time)
        trimmed_list = [word for word in self.wordlist if len(word) <= cipher_len]
        prepend_list = ['']

        last_num_processed = 0
        results = []

        for j in word_ranges:
            log.debug(f"Starting word range {j}")

            for i in range(j - last_num_processed):
                word_scores = []
                for prepend in prepend_list:
                    for word in trimmed_list:
                        mod_word = (prepend + delimiter + word).strip()
                        xor_result = xor_buffs((bytes(mod_word, 'utf-8') + b'\x00' * cipher_len)[:cipher_len], two_time)[:len(two_time)]
                        analysis = self.analyzer.analyze(xor_result)
                        word_scores.append((mod_word, analysis / (len(word) ** 2)))

                prepend_list = [word for word, _ in sorted(word_scores, key=lambda score: score[1], reverse=True)[:10 ** (i + 1 + last_num_processed)]]
            last_num_processed = j

            results.append(sorted(prepend_list, key=lambda word: self.analyzer.analyze(xor_buffs((bytes(word, 'utf-8') + b'\x00' * cipher_len)[:cipher_len], two_time)), reverse=True)[:10])
            
        return results


from samson.utilities.manipulation import left_rotate
from samson.utilities.bytes import Bytes


def initialize_sbox():
    p = 1
    q = 1

    sbox = [None] * 256
    first_round = True

    while first_round or p != 1:
        # Multiply p by 3
        p = p ^ (p << 1) ^ (0x1B if p & 0x80 else 0)
        p &= 0xFF
        
        # Divide q by 3
        q ^= (q << 1) & 0xFF
        q ^= (q << 2) & 0xFF
        q ^= (q << 4) & 0xFF
        q ^= 0x09 if q & 0x80 else 0
        q &= 0xFF

        # Compute the affine transformation
        xformed = q ^ left_rotate(q, 1, 8) ^ left_rotate(q, 2, 8) ^ left_rotate(q, 3, 8) ^ left_rotate(q, 4, 8)
        xformed &= 0xFF

        sbox[p] = xformed ^ 0x63
        first_round = False

    # Zero has no inverse and must be set manually
    sbox[0] = 0x63

    return sbox



def invert_sbox(sbox):
    inv_sbox = [None] * 256
    for i in range(256):
        inv_sbox[sbox[i]] = i

    return inv_sbox


RCON = [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1B, 0x36, 0x6C, 0xD8, 0xAB, 0x4D, 0x9A, 0x2F, 0x5E, 0xBC, 0x63, 0xC6, 0x97, 35, 0x6A, 0xD4, 0xB3, 0x7D, 0xFA, 0xEF, 0xC5]
SBOX = initialize_sbox()
INV_SBOX = invert_sbox(SBOX)
MIX_MATRIX = [2, 3, 1, 1, 1, 2, 3, 1, 1, 1, 2, 3, 3, 1, 1, 2]
INV_MIX_MATRIX = [14, 11, 13, 9, 9, 14, 11, 13, 13, 9, 14, 11, 11, 13, 9, 14]

SHIFT_ROW_OFFSETS = [
    *[[0, 1, 2, 3]] * 3,
    [0, 1, 2, 4],
    [0, 1, 3, 4]
]

NUM_ROUNDS = [
    [10, 12, 14],
    [12, 12, 14],
    [14, 14, 14]
]

# https://csrc.nist.gov/csrc/media/projects/cryptographic-standards-and-guidelines/documents/aes-development/rijndael-ammended.pdf
class Rijndael(object):
    """
    Underlying cipher of AES.

    Structure: Substitution–permutation network
    Key size: 128, 160, 192, 224, 256 bits
    Block size: 128, 160, 192, 224, 256 bits
    """

    def __init__(self, key: bytes, block_size: int=128):
        """
        Parameters:
            key      (bytes): Bytes-like object to key the cipher.
            block_size (int): The desired block size in bits.
        """
        key = Bytes.wrap(key)
        if not (len(key) * 8) in range(128, 257, 32):
            raise Exception("Invalid key size! Must be between 128 and 256 and a multiple of 32")
        
        if not block_size in range(128, 257, 32):
            raise Exception("Invalid 'block_size'! Must be between 128 and 256 and a multiple of 32")

        
        self.key = key
        self.block_size = block_size

        self._chunk_size = self.block_size // 32
        round_keys = self.key_schedule()
        self.round_keys = [Bytes(b''.join(round_keys[i:i + self._chunk_size])) for i in range(0, len(round_keys), self._chunk_size)]
        
        Nk = len(self.key) // 4
        Nb = self._chunk_size 
        self.num_rounds = NUM_ROUNDS[(Nk - 4) // 2][(Nb - 4) // 2] + 1


    def __repr__(self):
        return f"<Rijndael: key={self.key}, key_size={len(self.key) * 8}, block_size={self.block_size}>"

    def __str__(self):
        return self.__repr__()


    # https://en.wikipedia.org/wiki/Rijndael_key_schedule
    def key_schedule(self):
        N = len(self.key) // 4
        K = self.key.chunk(4)
        R = max(N, self._chunk_size) + 7

        W = []

        for i in range(self._chunk_size*R):
            if i < N:
                W_i = K[i]
            elif i % N == 0:
                W_i = W[i - N] ^ Bytes([SBOX[byte] for byte in W[i - 1].lrot(8)]) ^ bytes([RCON[i // N - 1], *[0] * 3])
            elif N > 6 and i % N == 4:
                W_i = W[i - N] ^ Bytes([SBOX[byte] for byte in W[i - 1]])
            else:
                W_i = W[i - N] ^ W[i - 1]

            W.append(W_i)

        return W



    def yield_encrypt(self, plaintext):
        state_matrix = Bytes.wrap(plaintext).transpose(4)

        for i in range(self.num_rounds):
            round_key = self.round_keys[i].transpose(4)
            if i == 0:
                state_matrix ^= round_key
            elif i < (self.num_rounds - 1):
                state_matrix = Bytes([SBOX[byte] for byte in state_matrix])
                state_matrix = self.shift_rows(state_matrix)
                state_matrix = Bytes(self.mix_columns(state_matrix, MIX_MATRIX))
                state_matrix ^= round_key
            else:
                state_matrix = Bytes([SBOX[byte] for byte in state_matrix])
                state_matrix = self.shift_rows(state_matrix)
                state_matrix ^= round_key
            
            yield state_matrix.transpose(self._chunk_size)


    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        return list(self.yield_encrypt(plaintext))[-1]



    def yield_decrypt(self, ciphertext):
        state_matrix = Bytes.wrap(ciphertext).transpose(4)

        reversed_round_keys = self.round_keys[::-1]

        for i in range(self.num_rounds):
            round_key = reversed_round_keys[i].transpose(4)
            if i == 0:
                state_matrix ^= round_key
            elif i < (self.num_rounds - 1):
                state_matrix = self.inv_shift_rows(state_matrix)
                state_matrix = Bytes([INV_SBOX[byte] for byte in state_matrix])
                state_matrix ^= round_key
                state_matrix = Bytes(self.mix_columns(state_matrix, INV_MIX_MATRIX))
            else:
                state_matrix = self.inv_shift_rows(state_matrix)
                state_matrix = Bytes([INV_SBOX[byte] for byte in state_matrix])
                state_matrix ^= round_key

            yield state_matrix.transpose(self._chunk_size)



    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        return list(self.yield_decrypt(ciphertext))[-1]


    def shift_rows(self, state_matrix):
        offsets = SHIFT_ROW_OFFSETS[(self._chunk_size) - 4]
        return b''.join([row.lrot(offsets[j] * 8) for j, row in enumerate(state_matrix.chunk(self._chunk_size))])


    def inv_shift_rows(self, state_matrix):
        offsets = SHIFT_ROW_OFFSETS[(self._chunk_size) - 4]
        return b''.join([row.rrot(offsets[j] * 8) for j, row in enumerate(state_matrix.chunk(self._chunk_size))])


    # https://en.wikipedia.org/wiki/Rijndael_MixColumns
    def _gmul(self, a, b):
        p = 0

        for _ in range(8):
            if (b & 1) != 0:
                p ^= a
            
            hi_bi_set = (a & 0x80) != 0
            a <<=1

            if hi_bi_set:
                a ^= 0x1B
            
            b >>= 1

        return p



    def mix_columns(self, state_matrix, mix_matrix):
        new_state = [None] * len(state_matrix)
        c0, c1, c2, c3 = [self._chunk_size * i for i in range(4)]

        for i in range(self._chunk_size):
            new_state[i + c0] = (self._gmul(mix_matrix[0], state_matrix[i + c0]) ^ self._gmul(mix_matrix[1], state_matrix[i + c1]) ^ self._gmul(mix_matrix[2], state_matrix[i + c2]) ^ self._gmul(mix_matrix[3], state_matrix[i + c3])) & 0xFF
            new_state[i + c1] = (self._gmul(mix_matrix[4], state_matrix[i + c0]) ^ self._gmul(mix_matrix[5], state_matrix[i + c1]) ^ self._gmul(mix_matrix[6], state_matrix[i + c2]) ^ self._gmul(mix_matrix[7], state_matrix[i + c3])) & 0xFF
            new_state[i + c2] = (self._gmul(mix_matrix[8], state_matrix[i + c0]) ^ self._gmul(mix_matrix[9], state_matrix[i + c1]) ^ self._gmul(mix_matrix[10], state_matrix[i + c2]) ^ self._gmul(mix_matrix[11], state_matrix[i + c3])) & 0xFF
            new_state[i + c3] = (self._gmul(mix_matrix[12], state_matrix[i + c0]) ^ self._gmul(mix_matrix[13], state_matrix[i + c1]) ^ self._gmul(mix_matrix[14], state_matrix[i + c2]) ^ self._gmul(mix_matrix[15], state_matrix[i + c3])) & 0xFF
        
        return new_state


from sympy.abc import x
from sympy import ZZ, Poly, GF, invert
from sympy.polys.polyerrors import NotInvertible
from samson.utilities.general import shuffle
from samson.utilities.math import is_power_of_two
from samson.utilities.bytes import Bytes
from sympy.ntheory import isprime
import math


def encode_bytes(in_arr: list) -> Bytes:
    """
    Encodes a list of numbers as Bytes.

    Parameters:
        in_arr (list): List to encode.

    Returns:
        Bytes: List encoded as Bytes.
    """
    return Bytes([num % 256 for num in in_arr])



def decode_bytes(in_bytes: bytes) -> list:
    """
    Decodes bytes into a list.

    Parameters:
        in_bytes (bytes): Bytes to decode.
    
    Returns:
        list: List of integer representations of the bytes.
    """
    return [num - 256 if num > 127 else num for num in in_bytes]



def rand_poly(length: int, len_non_zeroes: int, neg_ones_mod: int=0) -> Poly:
    """
    Generates a random polynomial.

    Parameters:
        length         (int): Desired length of polynomial.
        len_non_zeroes (int): Number of non-zero values in polynomial.
        neg_ones_mod   (int): Modifier that reduces the number of -1 coefficients.
    
    Returns:
        Poly: Random polynomial.
    """
    poly_arr =  [0]  * ((length - len_non_zeroes * 2) + neg_ones_mod)
    poly_arr += [1]  * len_non_zeroes
    poly_arr += [-1] * (len_non_zeroes - neg_ones_mod)

    shuffle(poly_arr)

    return Poly(poly_arr, x).set_domain(ZZ)



def invert_poly(f_poly: Poly, R_poly: Poly, p: int) -> Poly:
    """
    Inverts a polynomial `f_poly` over `R_poly` in GF(p).

    Parameters:
        f_poly (Poly): Polynomial to be inverted.
        R_poly (Poly): Polynomial to be inverted _over_.
        p       (int): Integer modulus.
    
    Returns:
        Poly: Inverted polynomial.
    """
    if isprime(p):
        inv_poly = invert(f_poly, R_poly, domain=GF(p))

    elif is_power_of_two(p):
        inv_poly = invert(f_poly, R_poly, domain=GF(2))

        for _ in range(int(math.log(p, 2))):
            inv_poly = (2 * inv_poly) - (f_poly * (inv_poly ** 2))
            inv_poly = (inv_poly % R_poly).trunc(p)
    else:
        raise Exception(f"Polynomial not invertible in Z_{p}. NTRU: p and q must be prime or power of two.")
    
    return inv_poly



# https://en.wikipedia.org/wiki/NTRUEncrypt
class NTRU(object):
    """
    Nth-degree TRUncated polynomial ring
    """

    def __init__(self, N: int, p: int=3, q: int=128, f_poly: Poly=None, g_poly: Poly=None):
        """
        Parameters:
            N       (int): Polynomial degree/modulus.
            p       (int): Small modulus.
            q       (int): Large modulus.
            f_poly (Poly): F-polynomial of private key.
            g_poly (Poly): G-polynomial of private key.
        """
        self.N = N
        self.p = p
        self.q = q
        self.R_poly = Poly(x**N - 1, x).set_domain(ZZ)

        self.f_poly = f_poly
        self.g_poly = g_poly
        self.h_poly = None

        priv_not_specified = [poly is None for poly in [f_poly, g_poly]]

        # Generate random keys
        if all(priv_not_specified):
            self.generate_random_keys()

        # Tried to specify only part of private key
        elif any(priv_not_specified):
            raise Exception("Must provide ALL values for private key: f_poly, g_poly")

        # Specified private key, but not public key
        else:
            self.generate_public_key()



    def __repr__(self):
        return f"<NTRU: N={self.N}, p={self.p}, q={self.q}, f_poly={self.f_poly}, g_poly={self.g_poly}, h_poly={self.h_poly}>"

    def __str__(self):
        return self.__repr__()


    
    def generate_random_keys(self):
        """
        Generates random private and public keys.
        """
        self.g_poly = rand_poly(self.N, int(math.sqrt(self.q)))

        inversion_successful = False
        while not inversion_successful:
            try:
                self.f_poly = rand_poly(self.N, self.N // 3, neg_ones_mod=1)

                self.generate_public_key()
                inversion_successful = True
            except NotInvertible as _:
                pass
    


    def generate_public_key(self):
        """
        Attempts to find the public key for the current private key. May throw `NotInvertible`.
        """
        self.f_p_poly = invert_poly(self.f_poly, self.R_poly, self.p)
        self.f_q_poly = invert_poly(self.f_poly, self.R_poly, self.q)

        p_f_q_poly = (self.p * self.f_q_poly).trunc(self.q)
        pfq_trunc = (p_f_q_poly * self.g_poly).trunc(self.q)

        self.h_poly = (pfq_trunc % self.R_poly).trunc(self.q)



    def encrypt(self, plaintext: bytes, random_poly: Poly=None) -> Bytes:
        """
        Encrpyts `plaintext`.

        Parameters:
            plaintext  (bytes): Plaintext to encrypt.
            random_poly (Poly): (Optional) The random polynomial used in encryption.

        Returns:
            Bytes: Encrypted ciphertext.
        """
        random_poly = random_poly or rand_poly(self.N, int(math.sqrt(self.q)))

        # Convert plaintext into polynomial
        pt_poly = Poly([int(bit) for bit in bin(int.from_bytes(plaintext, 'big'))[2:].zfill(len(plaintext) * 8)][::-1], x).set_domain(ZZ)
        rhm = (random_poly * self.h_poly).trunc(self.q) + pt_poly
        ct_poly = (rhm % self.R_poly).trunc(self.q)

        return encode_bytes(ct_poly.all_coeffs()[::-1])
    


    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext` into plaintext.

        Parameters:
            ciphertext (bytes): Ciphertext.
        
        Returns:
            Bytes: Decrypted plaintext.
        """
        # Convert ciphertext into polynomial
        ct_poly = decode_bytes(ciphertext)
        msg_poly = Poly(ct_poly[::-1], x).set_domain(ZZ)

        a_poly = ((self.f_poly * msg_poly) % self.R_poly).trunc(self.q)
        b_poly = a_poly.trunc(self.p)

        pt_poly = ((self.f_p_poly * b_poly) % self.R_poly).trunc(self.p)

        bit_length = math.ceil(len(pt_poly.all_coeffs()) / 8) * 8
        pt_bitstring = ''.join([str(bit) for bit in pt_poly.all_coeffs()]).zfill(bit_length)[::-1]
        return Bytes(int(pt_bitstring, 2)).zfill(bit_length // 8)


from samson.protocols.diffie_hellman import DiffieHellman
from samson.utilities.bytes import Bytes
from samson.utilities.math import mod_inv

class ElGamal(object):
    """
    ElGamal public key encryption
    """

    def __init__(self, g: int=2, p: int=DiffieHellman.MODP_2048, key: int=None):
        """
        Parameters:
            g   (int): Generator.
            p   (int): Prime modulus.
            key (int): Key.
        """
        self.key = key or Bytes.random().int()
        self.g = g
        self.p = p
        self.pub = pow(self.g, self.key, self.p)
    
    
    def __repr__(self):
        return f"<ElGamal: key={self.key}, g={self.g}, p={self.p}, pub={self.pub}>"


    def __str__(self):
        return self.__repr__()


    # https://en.wikipedia.org/wiki/ElGamal_encryption
    def encrypt(self, plaintext: bytes, k: int=None) -> (int, int):
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Message to encrypt.
            k           (int): (Optional) Ephemeral key.
        
        Returns:
            (int, int): Formatted as (ephemeral key, ciphertext).
        """
        K_e = k or max(1, Bytes.random().int() % self.p)
        c_1 = pow(self.g, K_e, self.p)
        s = pow(self.pub, K_e, self.p)
        plaintext = Bytes.wrap(plaintext)
        return c_1, (s * plaintext.int()) % self.p
    


    def decrypt(self, key_and_ciphertext: (int ,int)) -> Bytes:
        """
        Decrypts `key_and_ciphertext`.

        Parameters:
            key_and_ciphertext ((int, int)): Ephemeral key and ciphertext.
        
        Returns:
            Bytes: Plaintext.
        """
        c_1, ciphertext = key_and_ciphertext
        s = pow(c_1, self.key, self.p)
        return Bytes((mod_inv(s, self.p) * ciphertext) % self.p)


from samson.utilities.manipulation import get_blocks
from samson.utilities.bytes import Bytes
from samson.oracles.stateless_block_encryption_oracle import StatelessBlockEncryptionOracle
import struct

import logging
log = logging.getLogger(__name__)

class ECBPrependAttack(object):
    """
    Performs a plaintext recovery attack.

    By prepending data to the secret, we can take advantage of ECB's statelessness and iteratively build
    ciphertext blocks that match the secret's ciphertext blocks.

    Conditions:
        * ECB is being used
        * The user has access to an oracle that accepts arbitrary plaintext and returns the ciphertext
        * The user's input is prepended to the secret plaintext
    """

    def __init__(self, oracle: StatelessBlockEncryptionOracle):
        """
        Parameters:
            oracle (StatelessBlockEncryptionOracle): An oracle that takes in plaintext and returns the ciphertext.
        """
        self.oracle = oracle


    def execute(self) -> Bytes:
        """
        Executes the attack.

        Parameters:
            unpad (bool): Whether or not to PKCS7 unpad the result.
        
        Returns:
            Bytes: The recovered plaintext.
        """
        baseline = len(self.oracle.encrypt(b''))
        block_size = self.oracle.find_block_size()

        plaintexts = []
        for curr_block in range(baseline // block_size):
            log.debug("Starting iteration {}".format(curr_block))
            
            plaintext = b''
            for curr_byte in range(block_size):
                if curr_block == 0:
                    payload = ('A' * (block_size - (curr_byte + 1))).encode()
                else:
                    payload = plaintexts[-1][curr_byte + 1:]

                one_byte_short = get_blocks(self.oracle.encrypt(payload), block_size=block_size)[curr_block]

                for i in range(256):
                    curr_byte = struct.pack('B', i)
                    ciphertext = self.oracle.encrypt(payload + plaintext + curr_byte)

                    # We're always editing the first block to look like block 'curr_block'
                    if get_blocks(ciphertext, block_size=block_size)[0] == one_byte_short:
                        plaintext += curr_byte
                        break

            plaintexts.append(plaintext)
        return Bytes(b''.join(plaintexts))



from samson.utilities.manipulation import right_rotate
from types import FunctionType

#V{output_size}_{state_size}_{transform}[.._{n_transforms}]
def V32_64_XSH_RR(x, mult, inc):
    count = x >> 59
    state = (x * mult + inc) & 0xFFFFFFFFFFFFFFFF
    x ^= x >> 18
    return state, right_rotate((x >> 27) & 0xFFFFFFFF, count)



# https://en.wikipedia.org/wiki/Permuted_congruential_generator
class PCG(object):
    def __init__(self, seed: int, multiplier: int, increment: int, variant: FunctionType=V32_64_XSH_RR):
        """
        Parameters:
            seed       (int): Initial value.
            multiplier (int): Multiplier.
            increment  (int): Increment.
            variant   (func): PCG variant. Function that takes in (state, multiplier, increment) and returns (new_state, output).
        """
        self.state = seed
        self.multiplier = multiplier
        self.increment = increment
        self.variant = variant


    def __repr__(self):
        return f"<PCG: state={self.state}, multiplier={self.multiplier}, increment={self.increment}, variant={self.variant}>"

    def __str__(self):
        return self.__repr__()


    
    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.state, result = self.variant(self.state, self.multiplier, self.increment)
        return result


from samson.block_ciphers.rijndael import SBOX as RIJ_SBOX
from samson.utilities.bytes import Bytes

SQ = [
    0x25,0x24,0x73,0x67,0xD7,0xAE,0x5C,0x30,0xA4,0xEE,0x6E,0xCB,0x7D,0xB5,0x82,0xDB,
    0xE4,0x8E,0x48,0x49,0x4F,0x5D,0x6A,0x78,0x70,0x88,0xE8,0x5F,0x5E,0x84,0x65,0xE2,
    0xD8,0xE9,0xCC,0xED,0x40,0x2F,0x11,0x28,0x57,0xD2,0xAC,0xE3,0x4A,0x15,0x1B,0xB9,
    0xB2,0x80,0x85,0xA6,0x2E,0x02,0x47,0x29,0x07,0x4B,0x0E,0xC1,0x51,0xAA,0x89,0xD4,
    0xCA,0x01,0x46,0xB3,0xEF,0xDD,0x44,0x7B,0xC2,0x7F,0xBE,0xC3,0x9F,0x20,0x4C,0x64,
    0x83,0xA2,0x68,0x42,0x13,0xB4,0x41,0xCD,0xBA,0xC6,0xBB,0x6D,0x4D,0x71,0x21,0xF4,
    0x8D,0xB0,0xE5,0x93,0xFE,0x8F,0xE6,0xCF,0x43,0x45,0x31,0x22,0x37,0x36,0x96,0xFA,
    0xBC,0x0F,0x08,0x52,0x1D,0x55,0x1A,0xC5,0x4E,0x23,0x69,0x7A,0x92,0xFF,0x5B,0x5A,
    0xEB,0x9A,0x1C,0xA9,0xD1,0x7E,0x0D,0xFC,0x50,0x8A,0xB6,0x62,0xF5,0x0A,0xF8,0xDC,
    0x03,0x3C,0x0C,0x39,0xF1,0xB8,0xF3,0x3D,0xF2,0xD5,0x97,0x66,0x81,0x32,0xA0,0x00,
    0x06,0xCE,0xF6,0xEA,0xB7,0x17,0xF7,0x8C,0x79,0xD6,0xA7,0xBF,0x8B,0x3F,0x1F,0x53,
    0x63,0x75,0x35,0x2C,0x60,0xFD,0x27,0xD3,0x94,0xA5,0x7C,0xA1,0x05,0x58,0x2D,0xBD,
    0xD9,0xC7,0xAF,0x6B,0x54,0x0B,0xE0,0x38,0x04,0xC8,0x9D,0xE7,0x14,0xB1,0x87,0x9C,
    0xDF,0x6F,0xF9,0xDA,0x2A,0xC4,0x59,0x16,0x74,0x91,0xAB,0x26,0x61,0x76,0x34,0x2B,
    0xAD,0x99,0xFB,0x72,0xEC,0x33,0x12,0xDE,0x98,0x3B,0xC0,0x9B,0x3E,0x18,0x10,0x3A,
    0x56,0xE1,0x77,0xC9,0x1E,0x9E,0x95,0xA3,0x90,0x19,0xA8,0x6C,0x09,0xD0,0xF0,0x86
]

# https://www.gsma.com/aboutus/wp-content/uploads/2014/12/snow3gspec.pdf
# https://github.com/mitshell/CryptoMobile/blob/master/C_alg/SNOW_3G.c
class SNOW3G(object):
    """
    SNOW3G stream cipher

    Used in 4G LTE encryption.
    """

    def __init__(self, key: bytes, iv: bytes):
        """
        Parameters:
            key (bytes): Key (128 or 256 bits).
            iv  (bytes): Initialization vector (16 bytes).
        """
        self.key = Bytes.wrap(key)
        self.iv = Bytes.wrap(iv)

        k0, k1, k2, k3 = [chunk.to_int() for chunk in self.key.chunk(4)]
        iv0, iv1, iv2, iv3 = [chunk.to_int() for chunk in self.iv.chunk(4)]

        s = [None] * 16
        s[15] = k3 ^ iv0
        s[14] = k2
        s[13] = k1
        s[12] = k0 ^ iv1
        s[11] = k3 ^ 0xFFFFFFFF
        s[10] = k2 ^ 0xFFFFFFFF ^ iv2
        s[9] = k1 ^ 0xFFFFFFFF ^ iv3
        s[8] = k0 ^ 0xFFFFFFFF
        s[7] = k3
        s[6] = k2
        s[5] = k1
        s[4] = k0
        s[3] = k3 ^ 0xFFFFFFFF
        s[2] = k2 ^ 0xFFFFFFFF
        s[1] = k1 ^ 0xFFFFFFFF
        s[0] = k0 ^ 0xFFFFFFFF

        self.s = s

        self.R1 = 0
        self.R2 = 0
        self.R3 = 0

        for _ in range(32):
            F = self.clock_FSM()
            self.clock_lfsr(F)
    


    def __repr__(self):
        return f"<SNOW3G: key={self.key}, iv={self.iv}, s={self.s}, R1={self.self.R1}, R2={self.self.R2}, R3={self.self.R3}>"

    def __str__(self):
        return self.__repr__()



    def MULx(self, V: int, c: int) -> int:
        if V >> 7:
            return ((V << 1) % 256) ^ c
        else:
            return V << 1


    def MULa(self, c: int) -> int:
        return (self.MULxPOW(c, 23, 0xA9) << 24) + (self.MULxPOW(c, 245, 0xA9) << 16) + (self.MULxPOW(c, 48, 0xA9) << 8) + self.MULxPOW(c, 239, 0xA9)

    
    def DIVa(self, c: int) -> int:
        return (self.MULxPOW(c, 16, 0xA9) << 24) + (self.MULxPOW(c, 39, 0xA9) << 16) + (self.MULxPOW(c, 6, 0xA9) << 8) + self.MULxPOW(c, 64, 0xA9)


    def MULxPOW(self, V: int, i: int, c: int) -> int:
        if i == 0:
            return V
        else:
            return self.MULx(self.MULxPOW(V, i - 1, c), c)
    


    def S1(self, w: int) -> Bytes:
        return self._perform_sbox_transform(Bytes.wrap(w).zfill(4), RIJ_SBOX, 0x1B, True)


    def S2(self, w: int) -> Bytes:
        return self._perform_sbox_transform(Bytes.wrap(w).zfill(4), SQ, 0x69)


    def _perform_sbox_transform(self, w: bytes, sbox: list, val: int, s2=False) -> Bytes:
        sqw0, sqw1, sqw2, sqw3 = [sbox[w_i] for w_i in w]

        r0 = self.MULx(sqw0, val) ^ sqw1 ^ sqw2 ^ self.MULx(sqw3, val) ^ sqw3
        r1 = self.MULx(sqw0, val) ^ sqw0 ^ self.MULx(sqw1, val)  ^ sqw2 ^ sqw3
        r2 = sqw0 ^ self.MULx(sqw1, val) ^ sqw1  ^ self.MULx(sqw2, val) ^ sqw3
        r3 = sqw0 ^ sqw1 ^ self.MULx(sqw2, val) ^ sqw2 ^ self.MULx(sqw3, val)

        return Bytes([r0, r1, r2, r3]).to_int()
    

    def clock_FSM(self) -> int:
        """
        Used internally. Clocks the internal FSM.

        Returns:
            int: Next value of `F`.
        """
        F = ((self.s[15] + self.R1) % (2 ** 32)) ^ self.R2
        r = (self.R2 + (self.R3 ^ self.s[5])) % (2 ** 32)
        self.R3 = self.S2(self.R2)
        self.R2 = self.S1(self.R1)
        self.R1 = r

        return F


    def clock_lfsr(self, F: int=None):
        """
        Used internally. Clocks the LFSR and possibly combines with `F`.

        Parameters:
            F (int): Value of `F` to be clocked with.
        """
        v = ((self.s[0] << 8) & 0xFFFFFF00) ^ self.MULa((self.s[0] >> 24) & 0xFF) ^ self.s[2] ^ ((self.s[11] >> 8) & 0x00FFFFFF) ^ self.DIVa(self.s[11] & 0xFF)
        if F:
            v^= F

        for i in range(15):
            self.s[i] = self.s[i + 1]

        self.s[15] = v



    def generate(self, length: int) -> Bytes:
        """
        Generates `length` of keystream.

        Parameters:
            length (int): Desired length of keystream in bytes.
        
        Returns:
            Bytes: Keystream.
        """
        _F = self.clock_FSM()
        self.clock_lfsr()

        ks = []
        
        for _ in range(length):
            F = self.clock_FSM()
            ks.append(F ^ self.s[0])
            self.clock_lfsr()

        return sum([Bytes(i).zfill(4) for i in ks])


from math import log10
import os

# http://practicalcryptography.com/cryptanalysis/text-characterisation/word-statistics-fitness-measure/
class ViterbiDecoder(object):
    """
    Statistical model that decodes non-delimited English text into tokens using maximum likelihood metrics.
    """

    def __init__(self):
        self.Pw = {}
        with open(os.path.join(os.path.dirname(__file__), 'count_1w.txt'), 'r') as f:
            for line in f.readlines():
                key, count = line.split('\t') 

                self.Pw[key.upper()] = self.Pw.get(key.upper(), 0) + int(count)

        self.N = 1024908267229 ## Number of tokens


        # Calculate first order log probabilities
        for key in self.Pw.keys():
            self.Pw[key] = log10(float(self.Pw[key]) / self.N)


        # Get second order word model 
        self.Pw2 = {}
        with open(os.path.join(os.path.dirname(__file__), 'count_2w.txt'), 'r') as f:
            for line in f.readlines():
                key, count = line.split('\t') 

                self.Pw2[key.upper()] = self.Pw2.get(key.upper(), 0) + int(count)


        # Calculate second order log probabilities
        for key in self.Pw2.keys():
            word1, _word2 = key.split()

            if word1 not in self.Pw: 
                self.Pw2[key] = log10(float(self.Pw2[key]) / self.N)
            else: 
                self.Pw2[key] = log10(float(self.Pw2[key]) / self.N) - self.Pw[word1]


        # Precalculate the probabilities we assign to words not in our dict, L is length of word
        self.unseen = [log10(10. / (self.N * 10**L)) for L in range(50)]        
        
        


    def cPw(self, word: str, prev: str='<UNK>') -> float:
        """
        Calculates the conditional word probability.

        Parameters:
            word (str): Current word.
            prev (str): Previous word.
        
        Returns:
            float: Log probability of `word` based on `prev`.
        """
        second_order_word = (prev + ' ' + word)

        if word not in self.Pw: 
            return self.unseen[len(word)]

        elif second_order_word not in self.Pw2: 
            return self.Pw[word]

        else: 
            return self.Pw2[second_order_word]
    



    def score(self, text: str, max_word_len=20) -> (float, list):
        """
        Scores and tokenizes the text according to the maximum likelihood.

        Parameters:
            text        (str): Text to tokenize/decode.
            max_word_len(int): Maximum token length.
        
        Returns:
            (float, list): Most probable decoding as (score, token_list).
        """
        text = text.upper()

        prob = [[-99e99] * max_word_len for _ in range(len(text))]
        strs = [[''] * max_word_len for _ in range(len(text))]

        for j in range(max_word_len):
            prob[0][j] = self.cPw(text[:j+1])
            strs[0][j] = [text[:j+1]]

        for i in range(1,len(text)):
            for j in range(max_word_len):
                if i+j+1 > len(text): break
                candidates = [(prob[i-k-1][k] + self.cPw(text[i: i+j+1], strs[i-k-1][k][-1]),
                               strs[i-k-1][k] + [text[i: i+j+1]]) for k in range(min(i, max_word_len))]

                prob[i][j], strs[i][j] = max(candidates)
                
        ends = [(prob[-i-1][i], strs[-i-1][i]) for i in range(min(len(text), max_word_len))]
        return max(ends)


from samson.utilities.math import mod_inv

class AffineCipher(object):
    """
    Monoalphabetic substitution of the form `(ax + b) mod m`.

    The Caesar cipher is an AffineCipher with a=1.
    """

    def __init__(self, a: int, b: int, alphabet: str='ABCDEFGHIJKLMNOPQRSTUVWXYZ ,.'):
        """
        Parameters:
            a        (int): Multiplier.
            b        (int): Linear shift.
            alphabet (str): Alphabet (in order) to encrypt over. Input must also be in this alphabet.
        """
        self.a = a
        self.b = b
        self.alphabet = alphabet

        self.char_map = {}
        for i in range(len(alphabet)):
            self.char_map[alphabet[i]] = alphabet[(a*i+b)%len(alphabet)]


        inv_a = mod_inv(a, len(alphabet))
        self.inv_char_map = {}
        for i in range(len(alphabet)):
            self.inv_char_map[alphabet[i]] = alphabet[(inv_a*(i-b))%len(alphabet)]


    def __repr__(self):
        return "<Affine: a={}, b={}, alphabet={}>".format(self.a, self.b, self.alphabet)

    def __str__(self):
        return self.__repr__()


    def encrypt(self, plaintext: str) -> str:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (str): String to be encrypted.
        
        Returns:
            str: Resulting ciphertext.
        """
        primitive_len = len(self.alphabet[0])

        ciphertext = ''
        for i in range(0, len(plaintext), primitive_len):
            curr_symbol = plaintext[i:i + primitive_len]

            if curr_symbol in self.char_map:
                ciphertext += self.char_map[curr_symbol]
            else:
                ciphertext += curr_symbol

        return ciphertext


    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (str): Stringt to be decrypted.
        
        Returns:
            str: Resulting plaintext.
        """
        primitive_len = len(self.alphabet[0])

        plaintext = ''
        for i in range(0, len(ciphertext), primitive_len):
            curr_symbol = ciphertext[i:i + primitive_len]

            if curr_symbol in self.inv_char_map:
                plaintext += self.inv_char_map[curr_symbol]
            else:
                plaintext += curr_symbol

        return plaintext


from sympy import Poly
from samson.utilities.math import berlekamp_massey
from samson.utilities.encoding import poly_to_int

class GLFSR(object):
    """
    Galois linear-feedback shift register.
    """

    def __init__(self, seed: int, polynomial: Poly):
        """
        Parameters:
            seed        (int): Initial value.
            polynomial (Poly): Either a `sympy` `Poly` or an integer that represents the polynomal.
        """
        self.state = seed

        if type(polynomial) is Poly:
            polynomial = poly_to_int(polynomial)

        self.polynomial = polynomial
        self.mask = 1
        self.wrap_around_mask = 2 ** polynomial.bit_length() - 1
        self.state &= self.wrap_around_mask


        poly_mask = polynomial
        while poly_mask:
            if poly_mask & self.mask:
                poly_mask ^= self.mask
            
            if not poly_mask:
                break
            
            self.mask <<= 1


    def __repr__(self):
        return f"<GLFSR: state={self.state}, polynomial={self.polynomial}, mask={self.mask}>"
    

    def __str__(self):
        return self.__repr__()



    def clock(self):
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.state <<= 1
        self.state &= self.wrap_around_mask

        if self.state & self.mask:
            self.state ^= self.polynomial
            return 1

        else:
            return 0
        
        

    def generate(self) -> int:
        """
        Calls self.clock(). Here for interface uniformity
        
        Returns:
            int: Next psuedorandom output.
        """
        return self.clock()

    
    
    def reverse_clock(self, output: int):
        """
        Clocks the state in reverse given the previous output.

        Parameters:
            output (int): Previous output.
        """
        for item in output:
            if item:
                self.state ^= self.polynomial
            
            self.state >>=1
            self.state &= self.wrap_around_mask



    @staticmethod
    def crack(outputs: list):
        """
        Given a list of outputs, creates a GLFSR that generates the same sequence.

        Parameters:
            outputs (list): A list of outputs from the GLFSR (in order).
        
        Returns:
            GLFSR: GLFSR that generates the same sequence.
        """
        # Find minimum polynomial that represents the output
        poly = berlekamp_massey(outputs)

        # Create new LFSR and clock in reverse
        lfsr = GLFSR(0, poly)
        lfsr.reverse_clock(outputs[::-1])

        # Clock forward to synchronize with output
        [(lfsr.clock(), lfsr.state) for i in range(len(outputs))]

        return lfsr


start_exec = """from __future__ import division
from sympy import *
x, y, z, t = symbols('x y z t')
k, m, n = symbols('k m n', integer=True)
f, g, h = symbols('f g h', cls=Function)

from samson.all import *
from fastecdsa.curve import *

init_printing()

import logging
logging.basicConfig(format='%(asctime)s - %(name)s [%(levelname)s] %(message)s', level=logging.DEBUG)
"""

LOGO = """
                                                                
  /$$$$$$$  /$$$$$$  /$$$$$$/$$$$   /$$$$$$$  /$$$$$$  /$$$$$$$ 
 /$$_____/ |____  $$| $$_  $$_  $$ /$$_____/ /$$__  $$| $$__  $$
|  $$$$$$   /$$$$$$$| $$ \ $$ \ $$|  $$$$$$ | $$  \ $$| $$  \ $$
 \____  $$ /$$__  $$| $$ | $$ | $$ \____  $$| $$  | $$| $$  | $$
 /$$$$$$$/|  $$$$$$$| $$ | $$ | $$ /$$$$$$$/|  $$$$$$/| $$  | $$
|_______/  \_______/|__/ |__/ |__/|_______/  \______/ |__/  |__/
                                                                
                                                                
                                                                """

def start_repl():
    """
    Executes the samson REPL.
    """
    import IPython
    import sys
    from samson import VERSION

    banner = f"""
{LOGO}
    v{VERSION} -- https://github.com/wildcardcorp/samson

Python {sys.version}
IPython {IPython.__version__}
"""

    IPython.start_ipython(display_banner=False, exec_lines=[start_exec, f'print("""{banner}""")'])


from samson.utilities.bytes import Bytes
from types import FunctionType

class MatyasMeyerOseasConstruction(object):
    """
    Considered to be the dual of the Davies-Meyer construction. The Matyas-Meyer-Oseas construction is a
    one-way compression function built from a block cipher.
    """

    def __init__(self, initial_state: bytes, encryptor: FunctionType):
        """
        Parameters:
            initial_state (bytes): Bytes-like initial state that is the correct size for the underlying cipher.
            encryptor      (func): Function that takes in a plaintext and returns a ciphertext.
        """
        self.initial_state = initial_state
        self.encryptor = encryptor


    def __repr__(self):
        return f"<MatyasMeyerOseasConstruction initial_state={self.initial_state}, encryptor={self.encryptor}>"

    def __str__(self):
        return self.__repr__()


    def yield_state(self, message: bytes) -> Bytes:
        """
        Yields the intermediate, hashed states of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: Intermediate, hashed states.
        """
        block_size = len(self.initial_state)
        message = Bytes.wrap(message)

        message_blocks = message.chunk(block_size)
        last_state = self.initial_state

        for block in message_blocks:
            last_state ^= self.encryptor(last_state, block)
            yield last_state



    def hash(self, message: bytes) -> Bytes:
        """
        Yields the final, hashed state of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: Fully-hashed state.
        """
        final_state = [_ for _ in self.yield_state(message)][-1]
        return final_state


from samson.utilities.bytes import Bytes
from samson.utilities.encoding import poly_to_int, int_to_poly
from samson.utilities.manipulation import right_rotate, left_rotate

GF_MOD = 2**8 + 2**6 + 2**5 + 2**3 + 1
GF_MOD_POLY = int_to_poly(GF_MOD)

RS_MOD = 2**8 + 2**6 + 2**3 + 2**2 + 1
RS_MOD_POLY = int_to_poly(RS_MOD)

RHO = 0x01010101
ROUNDS = 16


MDS = [
    [ 0x01, 0xEF, 0x5B, 0x5B ],
    [ 0x5B, 0xEF, 0xEF, 0x01 ],
    [ 0xEF, 0x5B, 0x01, 0xEF ],
    [ 0xEF, 0x01, 0xEF, 0x5B ]
]

RS = [
    [ 0x01, 0xA4, 0x55, 0x87, 0x5A, 0x58, 0xDB, 0x9E ],
    [ 0xA4, 0x56, 0x82, 0xF3, 0x1E, 0xC6, 0x68, 0xE5 ],
    [ 0x02, 0xA1, 0xFC, 0xC1, 0x47, 0xAE, 0x3D, 0x19 ],
    [ 0xA4, 0x55, 0x87, 0x5A, 0x58, 0xDB, 0x9E, 0x03 ]
]

Q0 = [
    [ 0x8,0x1,0x7,0xD, 0x6,0xF,0x3,0x2, 0x0,0xB,0x5,0x9, 0xE,0xC,0xA,0x4 ],
    [ 0xE,0xC,0xB,0x8, 0x1,0x2,0x3,0x5, 0xF,0x4,0xA,0x6, 0x7,0x0,0x9,0xD ],
    [ 0xB,0xA,0x5,0xE, 0x6,0xD,0x9,0x0, 0xC,0x8,0xF,0x3, 0x2,0x4,0x7,0x1 ],
    [ 0xD,0x7,0xF,0x4, 0x1,0x2,0x6,0xE, 0x9,0xB,0x3,0x0, 0x8,0x5,0xC,0xA ]
]

Q1 = [
    [ 0x2,0x8,0xB,0xD, 0xF,0x7,0x6,0xE, 0x3,0x1,0x9,0x4, 0x0,0xA,0xC,0x5 ],
    [ 0x1,0xE,0x2,0xB, 0x4,0xC,0x3,0x7, 0x6,0xD,0xA,0x5, 0xF,0x9,0x0,0x8 ],
    [ 0x4,0xC,0x7,0x5, 0x1,0x6,0x9,0xA, 0x0,0xE,0xD,0x8, 0x2,0xB,0x3,0xF ],
    [ 0xB,0x9,0x5,0x1, 0xC,0x3,0xD,0xE, 0x6,0x4,0x7,0xF, 0x2,0x0,0x8,0xA ]
]

Q_BOXES = [Q0, Q1]


Q_ORD = [
    [ 1, 0, 1, 0 ],
    [ 0, 0, 1, 1 ],
    [ 0, 1, 0, 1 ],
    [ 1, 1, 0, 0 ],
    [ 1, 0, 0, 1 ]
]


def G(X, L, k):
    y = [_ for _ in Bytes(X).zfill(4)]

    for i in range(k - 1, -1, -1):
        for j in range(4):
            q_box = Q_BOXES[Q_ORD[i + 1][j]]
            y[j] = Q_PERMUTE(y[j], q_box) ^ L[i][j]
    

    for j in range(4):
        q_box = Q_BOXES[Q_ORD[0][j]]
        y[j] = Q_PERMUTE(y[j], q_box)
    

    z = MAT_MUL(MDS, y, GF_MOD_POLY)[::-1]
    return Bytes(z).int()



def PHT(a, b):
    return (a + b) % 2**32, (a + 2*b) % 2**32



def GF_MUL(a, b, poly_mod):
    poly_a = int_to_poly(a)
    poly_b = int_to_poly(b)
    return poly_to_int((poly_a * poly_b) % poly_mod)



def Q_PERMUTE(x, Q):
    a_0, b_0 = x // 16, x % 16
    a_1 = a_0 ^ b_0
    b_1 = (a_0 ^ right_rotate(b_0, 1, bits=4) ^ (a_0 * 8)) % 16

    a_2, b_2 = Q[0][a_1], Q[1][b_1]
    a_3 = a_2 ^ b_2
    b_3 = (a_2 ^ right_rotate(b_2, 1, bits=4) ^ (a_2 * 8)) % 16

    a_4, b_4 = Q[2][a_3], Q[3][b_3]
    return (b_4 * 16) + a_4



def MAT_MUL(mat, vec, mod):
    result = []
    for i in range(len(mat)):
        t = 0
        for j in range(len(vec)):
            t ^= GF_MUL(mat[i][j], vec[j], mod)
        
        result.append(t)
    
    return result


def EXPAND_KEY(M_e, M_o, k):
    K = []
    for i in range(ROUNDS + 4):
        A_i = G(2*i*RHO, M_e, k)
        B_i = G((2*i + 1) * RHO, M_o, k)
        B_i = left_rotate(B_i, 8)

        K.append((A_i + B_i) & 0xFFFFFFFF)
        K.append(left_rotate((A_i + 2*B_i) & 0xFFFFFFFF, 9))
    
    return K



# https://www.schneier.com/academic/paperfiles/paper-twofish-paper.pdf
class Twofish(object):
    """
    Structure: Feistel Network
    Key size: 128, 192, 256 bits
    Block size: 128 bits
    """

    def __init__(self, key: bytes):
        """
        Parameters:
            key (bytes): Bytes-like object to key the cipher.
        """
        self.key = Bytes.wrap(key)
        self.S = []
        self.K = None
        self.k = 0
        self._key_schedule()



    def __repr__(self):
        return f"<Twofish: key={self.key}, key_size={len(self.key) * 8}, S={self.S}, K={self.K}>"

    def __str__(self):
        return self.__repr__()



    def _key_schedule(self):
        M = self.key.chunk(4)
        self.k = k = len(M) // 2

        M_e = [M[i] for i in range(0, 2*k-1, 2)]
        M_o = [M[i] for i in range(1, 2*k,   2)]

        for i in range(k):
            vec = [byte for byte in M_e[i]] + [byte for byte in M_o[i]]
            self.S.append(MAT_MUL(RS, vec, RS_MOD_POLY))
        

        self.S.reverse()

        M_e_8 = [[_ for _ in chunk] for chunk in M_e]
        M_o_8 = [[_ for _ in chunk] for chunk in M_o]

        self.K = EXPAND_KEY(M_e_8, M_o_8, k)

    

    def F(self, R_0, R_1, i):
        R_0_little = int.from_bytes(int.to_bytes(R_0, 4, 'big'), 'little')
        R_1_little = int.from_bytes(int.to_bytes(left_rotate(R_1, 8), 4, 'big'), 'little')

        T_0 = G(R_0_little, self.S, self.k)
        T_1 = G(R_1_little, self.S, self.k)

        F_0 = (T_0 +   T_1 + self.K[2*i + 8]) & 0xFFFFFFFF
        F_1 = (T_0 + 2*T_1 + self.K[2*i + 9]) & 0xFFFFFFFF

        return F_0, F_1



    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        plaintext = Bytes.wrap(plaintext)
        pt_chunks = [chunk.zfill(4).int() for chunk in plaintext.chunk(4)]

        # Whitening
        R = [pt_chunks[i] ^ self.K[i] for i in range(len(pt_chunks))]

        for i in range(ROUNDS):
            FR_0, FR_1 = self.F(R[0], R[1], i)
            R = [
                R[0],
                R[1],
                right_rotate(R[2] ^ FR_0, 1),
                left_rotate(R[3], 1) ^ FR_1
            ]

            R[0], R[2] = R[2], R[0]
            R[1], R[3] = R[3], R[1]
        
        R = [R[(i+2) % 4] ^ self.K[i+4] for i in range(len(pt_chunks))]
        
        return Bytes(b''.join([int.to_bytes(r, 4, 'little') for r in R]))

    


    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        ciphertext = Bytes.wrap(ciphertext)
        ct_chunks = [chunk.zfill(4)[::-1].int() for chunk in ciphertext.chunk(4)]

        # Dewhitening
        R = [ct_chunks[i] ^ self.K[i+4] for i in range(len(ct_chunks))]

        for i in range(ROUNDS -1, -1, -1):
            FR_0, FR_1 = self.F(R[0], R[1], i)
            R = [
                R[0],
                R[1],
                left_rotate(R[2], 1)  ^ FR_0,
                right_rotate(R[3] ^ FR_1, 1)
            ]

            R[0], R[2] = R[2], R[0]
            R[1], R[3] = R[3], R[1]
        
        R = [R[(i+2) % 4] ^ self.K[i] for i in range(len(ct_chunks))]
        
        return Bytes(b''.join([int.to_bytes(r, 4, 'little') for r in R]))


from samson.oracles.encryption_oracle import EncryptionOracle
from samson.utilities.bytes import Bytes


class XORBitflippingAttack(object):
    """
    Performs a plaintext manipulation attack.

    CTR/stream ciphers will XOR the keystream with the ciphertext to decrypt the data. By injecting known plaintext,
    we can recover that block's keystream and XOR in our own plaintext. While this may seem extraneous given
    we can already inject a payload, this particular attack lends itself to bypassing validation functions.
    Imagine if an application encrypts a URL-encoded payload with its application key and gives it to us as way of keeping state.
    The developer assumes we can't modify the ciphertext without corrupting the string, so they trust the ciphertext
    when it's received.

    CBC will XOR the bitshift of the edited cipher block
    with the next blocks. To exploit this structure, we must
    craft a payload in reverse such that it creates our desired string.
    To do this, we need a known plaintext and a desired plaintext.
    We fill the targeted block with the known plaintext.
    We XOR our desired text, 'hiya;admin=true;', with the plaintext to find the "difference".
    Finally, we XOR the difference with the original cipher block.

    comment1=cooking
    %20MCs;userdata=
    aaaaaaaaaaaaaaaa
    ;comment2=%20lik
    e%20a%20pound%20
    of%20baconPPPPPP

    
    Conditions:
        * CBC, CTR, or a stream cipher is being used
        * The user has access to an oracle that allows encryption of arbitrary plaintext and returns the ciphertext.
    """

    def __init__(self, oracle: EncryptionOracle, block_size: int=16):
        """
        Parameters:
            oracle (EncryptionOracle): An oracle that takes in arbitrary plaintext and returns the ciphertext.
            block_size          (int): Block size of the underlying block cipher.
        """
        self.oracle = oracle
        self.block_size = block_size


    def execute(self, desired_injection: bytes, index: int=16) -> Bytes:
        """
        Executes the attack.

        Parameters:
            desired_injection (bytes): 
            index               (int): 
        
        Return:
            Bytes: The manipulated ciphertext.
        """
        payload = b'a' * self.block_size
        ciphertext = self.oracle.encrypt(payload)

        end_of_block = index + self.block_size
        edited_cipher = Bytes(ciphertext)
        edited_cipher[index:end_of_block] = edited_cipher[index:end_of_block] ^ desired_injection ^ payload
        return edited_cipher


import math
from types import FunctionType

import logging
log = logging.getLogger(__name__)


class NaiveMDCollider(object):
    """
    Merkle-Damgard collider that (naively) uses bruteforce to find collisions. Do not use for cryptographic
    strength algorithms.
    """

    def __init__(self, construction_func: FunctionType, output_size: int):
        """
        Parameters:
            construction_func (func): Function that takes in two bytes-like arguments
                                      and returns the next Merkle-Damgard state.

        Example:
        def construction_func(iv, message):
            return MerkleDamgardConstruction(iv, compressor, padder, output_size=hash_size).yield_state(message)


            output_size        (int): Size in bytes of the hash output.
        """
        self.construction_func = construction_func
        self.output_size = output_size


    
    @staticmethod
    def initialize_with_known_prefixes(prefixes: list, iv: bytes, construction_func: FunctionType, output_size: int):
        """
        Initializes a NaiveMDCollider using known prefixes. In Nostradamus-attack terms, the initial states
        you're trying to find collisions for.

        Parameters:
            prefixes          (list): List of bytes-like prefixes to build initial states from.
            iv               (bytes): Initial state of the Merkle-Damgard function (possibly None).
            construction_func (func): Function that takes in two bytes-like arguments
                                      and returns the next Merkle-Damgard state.
            output_size        (int): Size in bytes of the hash output.
        
        Returns:
            NaiveMDCollider: NaiveMDCollider initialized with known prefixes.
        """
        k = math.ceil(math.log(len(prefixes), 2))

        # Find the maximum length of padded prefixes
        padding_size = math.ceil(max([len(prefix) for prefix in prefixes]) / output_size) * output_size

        prefixes = [(prefix + (b'\x00' * padding_size))[:padding_size] for prefix in prefixes]
        hashed_prefixes = [list(construction_func(iv, prefix))[-1] for prefix in prefixes]

        return NaiveMDCollider(construction_func, output_size), k, prefixes, hashed_prefixes




    def find_collision(self, p1: bytes, p2: bytes) -> (bytes, bytes, bytes):
        """
        Finds a collision using bruteforce.

        Parameters:
            p1 (bytes): First sample.
            p2 (bytes): Second sample.
        
        Returns:
            (bytes, bytes, bytes): Tuple of bytes representing the collision as (p1_suffix, p2_suffix, self.hasher.hash(p1 + p1_suffix)).
        """
        input_for_p1 = b'\x00' * self.output_size
        state_to_collide = list(self.construction_func(p1, input_for_p1))[0]

        # Try a lot of values...
        for j in range(2**(self.output_size * 32)):
            attempt = int.to_bytes(j, self.output_size * 4, 'little')

            # We'll need to keep a state counter since we may not get a collision on the first chunk
            state_ctr = 0

            for state in self.construction_func(p2, attempt):
                state_ctr += 1

                if state == state_to_collide:
                    log.debug('Found collision for ({}, {})'.format(p1, p2))
                    return input_for_p1, attempt[:state_ctr * self.output_size], state_to_collide



from math import log
from samson.utilities.bytes import Bytes
from samson.utilities.manipulation import left_rotate
from samson.constructions.sponge_construction import SpongeConstruction

# https://github.com/ctz/keccak/blob/master/keccak.py

RC = [
  0x0000000000000001,   0x0000000000008082,   0x800000000000808A,   0x8000000080008000,
  0x000000000000808B,   0x0000000080000001,   0x8000000080008081,   0x8000000000008009,
  0x000000000000008A,   0x0000000000000088,   0x0000000080008009,   0x000000008000000A,
  0x000000008000808B,   0x800000000000008B,   0x8000000000008089,   0x8000000000008003,
  0x8000000000008002,   0x8000000000000080,   0x000000000000800A,   0x800000008000000A,
  0x8000000080008081,   0x8000000000008080,   0x0000000080000001,   0x8000000080008008
]


R = [
  [  0,  1, 62, 28, 27, ],
  [ 36, 44,  6, 55, 20, ],
  [  3, 10, 43, 25, 39, ],
  [ 41, 45, 15, 21,  8, ],
  [ 18,  2, 61, 56, 14, ]
]

# https://keccak.team/keccak_specs_summary.html
class Keccak(SpongeConstruction):
    """
    SHA3 winner based on the SpongeConstruction.
    """

    def __init__(self, r: int, c: int, digest_bit_size: int, auto_reset_state: bool=True):
        """
        Parameters:
            r                 (int): Bit-size of the sponge function.
            c                 (int): Sponge capacity.
            digest_bit_size   (int): Desired size of output.
            auto_reset_state (bool): Whether or not to reset the internal state before hashing.
        """
        super().__init__(self.keccak_f, self.pad, r, c)
        self.w = (r + c) // 25

        self.n = int(log(self.w, 2) * 2 + 12)
        self.digest_size = (digest_bit_size // 8)
        self.auto_reset_state = auto_reset_state

    

    def __repr__(self):
        return f"<Keccak r={self.r}, c={self.c}, n={self.n}, w={self.w}, digest_size={self.digest_size}, block_size={self.block_size}>"

    def __str__(self):
        return self.__repr__()
        

    def pad(self, in_bytes: bytes) -> bytes:
        bit_rate_bytes = (self.r + 7) // 8
        pad_len = (bit_rate_bytes - len(in_bytes)) % bit_rate_bytes

        if pad_len == 0:
            pad_len = bit_rate_bytes

        if pad_len == 1:
            return in_bytes + bytes([0x81])
        else:
            return in_bytes + bytes([0x01] + ([0] * (pad_len - 2)) + [0x80])


    def keccak_f(self, A):
        for i in range(self.n):
            A = self.round_func(A, RC[i])
        return A


    def round_func(self, A, rc):
        C = [0] * 5
        for x in range(5):
            C[x] = A[x][0] ^ A[x][1] ^ A[x][2] ^ A[x][3] ^ A[x][4]

        D = [0] * 5
        for x in range(5):
            D[x] = C[x-1] ^ left_rotate(C[(x+1) % 5], 1, 64)


        for x in range(5):
            for y in range(5):
                A[x][y] = A[x][y] ^ D[x]


        B = [[0] * 5 for _ in range(5)]
        for x in range(5):
            for y in range(5):
                B[y][(2*x + 3*y) % 5] = left_rotate(A[x][y], R[y][x], 64)


        for x in range(5):
            for y in range(5):
                A[x][y] = B[x][y] ^ ((~B[(x+1) % 5][y]) & B[(x+2) % 5][y])

        A[0][0] ^= rc

        return A
    


    def hash(self, message: bytes) -> Bytes:
        """
        Hashes the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: The hash digest.
        """
        if self.auto_reset_state:
            self.reset()
            
        self.absorb(Bytes.wrap(message))
        return sum(self.squeeze(self.digest_size))[:self.digest_size]


from samson.utilities.manipulation import left_rotate, get_blocks
from samson.utilities.bytes import Bytes
from samson.stream_ciphers.salsa import Salsa
from copy import deepcopy

# from samson.utilities
# https://en.wikipedia.org/wiki/Salsa20


def QUARTER_ROUND(a: int, b: int, c: int, d: int) -> (int, int, int, int):
    """
    Performs a quarter round of ChaCha.

    Parameters:
        a (int): ChaCha state variable.
        b (int): ChaCha state variable.
        c (int): ChaCha state variable.
        d (int): ChaCha state variable.
    
    Returns:
        (int, int, int, int): New values for (a, b, c, d).
    """
    a = (a + b) & 0xFFFFFFFF; d ^= a; d = left_rotate(d, 16)
    c = (c + d) & 0xFFFFFFFF; b ^= c; b = left_rotate(b, 12)
    a = (a + b) & 0xFFFFFFFF; d ^= a; d = left_rotate(d, 8)
    c = (c + d) & 0xFFFFFFFF; b ^= c; b = left_rotate(b, 7)
    return a, b, c, d


class ChaCha(Salsa):
    """
    ChaCha stream cipher

    Add-rotate-xor (ARX) structure.
    """
    
    def __init__(self, key: bytes, nonce: bytes, rounds: int=20, constant: bytes=b"expand 32-byte k"):
        """
        Parameters:
            key      (bytes): Key (128 or 256 bits).
            nonce    (bytes): Nonce (8 bytes).
            rounds     (int): Number of rounds to perform.
            constant (bytes): Constant used in generating the keystream (16 bytes).
        """
        self.key = key
        self.nonce = nonce
        self.rounds = rounds
        self.constant = constant



    def __repr__(self):
        return f"<ChaCha: key={self.key}, nonce={self.nonce}, rounds={self.rounds}, constant={self.constant}>"

    def __str__(self):
        return self.__repr__()


    def full_round(self, block_num: int, state: list=None) -> Bytes:
        """
        Performs a full round of ChaCha.

        Parameters:
            block_num (int): Current block number.
        
        Returns:
            Bytes: Keystream block.
        """
        ctr_bytes = int.to_bytes(block_num, 4, 'little')

        x = state or [
            *[int.from_bytes(block, 'little') for block in get_blocks(self.constant, 4)],
            *[int.from_bytes(block, 'little') for block in get_blocks(self.key, 4)],
                int.from_bytes(ctr_bytes, 'little'),
            *[int.from_bytes(block, 'little') for block in get_blocks(self.nonce, 4)]
        ]


        tmp = deepcopy(x)

        for _ in range(self.rounds // 2):
            # Odd round
            x[0], x[4], x[ 8], x[12] = QUARTER_ROUND(x[0], x[4], x[ 8], x[12])
            x[1], x[5], x[ 9], x[13] = QUARTER_ROUND(x[1], x[5], x[ 9], x[13])
            x[2], x[6], x[10], x[14] = QUARTER_ROUND(x[2], x[6], x[10], x[14])
            x[3], x[7], x[11], x[15] = QUARTER_ROUND(x[3], x[7], x[11], x[15])

            # Even round
            x[0], x[5], x[10], x[15] = QUARTER_ROUND(x[0], x[5], x[10], x[15])
            x[1], x[6], x[11], x[12] = QUARTER_ROUND(x[1], x[6], x[11], x[12])
            x[2], x[7], x[ 8], x[13] = QUARTER_ROUND(x[2], x[7], x[ 8], x[13])
            x[3], x[4], x[ 9], x[14] = QUARTER_ROUND(x[3], x[4], x[ 9], x[14])

        for i in range(16):
            x[i] += tmp[i]

        return Bytes(b''.join([int.to_bytes(state_int & 0xFFFFFFFF, 4, 'little') for state_int in x]), byteorder='little')


from .dghv import DGHV
from .dsa import DSA
from .ecdsa import ECDSA
from .eddsa import EdDSA
from .elgamal import ElGamal
from .merkle_hellman_knapsack import MerkleHellmanKnapsack
from .ntru import NTRU
from .paillier import Paillier
from .rsa import RSA


__all__ = ["DGHV", "DSA", "ECDSA", "EdDSA", "ElGamal", "MerkleHellmanKnapsack", "NTRU", "Paillier", "RSA"]


from samson.analyzers.all import *
from samson.attacks.all import *
from samson.auxiliary.all import *
from samson.block_ciphers.all import *
from samson.block_ciphers.modes.all import *
from samson.classical.all import *
from samson.constructions.all import *
from samson.hashes.all import *
from samson.kdfs.all import *
from samson.macs.all import *
from samson.oracles.all import *
from samson.padding.all import *
from samson.prngs.all import *
from samson.protocols.all import *
from samson.publickey.all import *
from samson.stream_ciphers.all import *
from samson.utilities.all import *


from .bitsliced_flfsr import BitslicedFLFSR
from .dual_ec import DualEC
from .flfsr import FLFSR
from .glfsr import GLFSR
from .lcg import LCG
from .lfg import LFG
from .mt19937 import MT19937
from .mwc1616 import MWC1616
from .pcg import PCG
from .xoroshiro import Xoroshiro
from .xorshift import Xorshift



__all__ = ["BitslicedFLFSR", "DualEC", "FLFSR", "GLFSR", "LCG", "LFG", "MT19937", "MWC1616", "PCG", "Xoroshiro", "Xorshift"]


from .gwo import GreyWolfOptimizer
from .hashclash_collider import HashClashCollider
from .markov_chain_handler import MarkovChainHandler
from .markov_state import MarkovState
from .naive_collider import NaiveMDCollider
from .token_list_handler import TokenListHandler
from .tokenizer import Tokenizer
from .viterbi_decoder import ViterbiDecoder


__all__ = ["GreyWolfOptimizer", "HashClashCollider", "MarkovChainHandler", "MarkovState", "NaiveMDCollider", "TokenListHandler", "Tokenizer", "ViterbiDecoder"]


from samson.utilities.manipulation import get_blocks
from samson.utilities.bytes import Bytes
from types import FunctionType
from copy import deepcopy


def md_pad(msg: bytes, fakeLen: int=None, byteorder: str='little', bit_size: int=64) -> bytes:
    length = fakeLen or len(msg)
    
    # Append the bit '1' to the message
    padding = b'\x80'
    byte_size = bit_size // 8

    # Append 0 <= k < 512 bits '0', so that the resulting message length (in bytes)
    # as congruent to 56 (mod 64)
    padding += b'\x00' * (((bit_size - byte_size) - (length + 1) % bit_size) % bit_size)

    # Append length of message (before pre-processing), in bits, as 64-bit big-endian integer
    message_bit_length = length * 8
    padding += message_bit_length.to_bytes(byte_size, byteorder=byteorder)
    return msg + padding


class MerkleDamgardConstruction(object):
    """
    An iterative construction for building collision-resistant cryptographic hash functions from collision-resistant
    one-way compression functions. Used in MD4, MD5, SHA1, SHA2, RIPEMD, and more.
    """

    def __init__(self, initial_state: bytes, compression_func: FunctionType, digest_size: int, block_size: int=64, endianness: str='big'):
        """
        Parameters:
            initial_state   (bytes): Bytes-like initial state that is the correct size for the underlying compression function.
            compression_func (func): One-way compression function. Takes in the state and returns the next.
            digest_size       (int): Resulting digest size. Should be the same size as the `initial_state`.
            block_size        (int): Size of the internal state.
            endianness        (str): Endianess of the internal state.
        """
        self.initial_state = deepcopy(initial_state)

        # Allows for direct use of class and subclass overriding simultaneously
        if compression_func:
            self.compression_func = compression_func

        self.digest_size = digest_size
        self.block_size = block_size
        self.endianness = endianness



    def __repr__(self):
        return f"<MerkleDamgardConstruction initial_state={self.initial_state}, compression_func={self.compression_func}, block_size={self.block_size}, pad_func={self.pad_func}>"

    def __str__(self):
        return self.__repr__()


    def pad_func(self, message: bytes) -> Bytes:
        """
        Pads the message with Merkle-Damgard padding.

        Parameters:
            message (bytes): Message to be padded.
        
        Returns:
            Bytes: Padded message.
        """
        return md_pad(message, None, self.endianness, bit_size=self.block_size)


    def yield_state(self, message: bytes) -> Bytes:
        """
        Yields the intermediate, hashed states of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: Intermediate, hashed states.
        """
        state = self.initial_state

        for block in get_blocks(self.pad_func(message), self.block_size):
            state = self.compression_func(block, state)
            yield state


    def hash(self, message: bytes) -> Bytes:
        """
        Yields the final, hashed state of the `message`.

        Parameters:
            message (bytes): Message to be hashed.
        
        Returns:
            Bytes: Fully-hashed state.
        """
        final_state = list(self.yield_state(message))[-1]
        return final_state

    

    def length_extension(self, observed_output: bytes, message: bytes, bytes_to_append: bytes, secret_len: int) -> (Bytes, Bytes):
        """
        Performs a length-extension attack.

        Unaware developers may create a keyed-MAC using a Merkle-Damgard construction and prepending a
        secret key to the input. However, such schemes are susceptible to length-extension attacks where an
        attacker can forge a valid HMAC with an arbitrary, appended message.

        Parameters:
            observed_output (bytes): The observed, valid, hash output.
            message         (bytes): Message originally hashed.
            bytes_to_append (bytes): Bytes to append to the end.
            secret_len        (int): The length of the secret.
        
        Returns:
            (Bytes, Bytes): Result formatted as (crafted input, forged hash).
        """
        glue = md_pad(message, len(message) + secret_len, self.endianness, bit_size=self.block_size)[len(message):]

        fake_len = secret_len + len(message) + len(glue) + len(bytes_to_append)

        new_hash_obj = deepcopy(self)
        new_hash_obj.initial_state = observed_output
        new_hash_obj.pad_func = lambda msg: md_pad(msg, fake_len, self.endianness, bit_size=self.block_size)

        return Bytes(message + glue + bytes_to_append), new_hash_obj.hash(bytes_to_append)


from .dh25519 import DH25519
from .diffie_hellman import DiffieHellman
from .dragonfly import Dragonfly
from .ecdhe import ECDHE
from .srp_client import SRPClient
from .srp_server import SRPServer


__all__ = ["DH25519", "DiffieHellman", "Dragonfly", "ECDHE", "SRPClient", "SRPServer"]


from samson.utilities.ecc import Curve25519, MontgomeryCurve
from samson.utilities.general import rand_bytes

class DH25519(object):
    """
    Elliptical curve Diffie-Hellman using Montgomery curves.
    """

    def __init__(self, key: int=None, base: int=None, curve: MontgomeryCurve=Curve25519()):
        """
        Parameters:
            key               (int): Secret key that will be clamped to the curve.
            base              (int): Base multiplier used in generating the challenge.
            curve (MontgomeryCurve): The curve used.
        """
        self.key = curve.clamp_to_curve(key or int.from_bytes(rand_bytes(32), 'big'))
        self.base = base or curve.U


    def __repr__(self):
        return f"<DH25519: key={self.key}, base={self.base}>"

    def __str__(self):
        return self.__repr__()


    def get_challenge(self) -> int:
        """
        Gets the challenge.

        Returns:
            int: The integer challenge representing an `x` value on the curve.
        """
        return self.key * self.base


    def derive_key(self, challenge: int) -> int:
        """
        Derives the shared key from the other instance's challenge.

        Parameters:
            challenge (int): The other instance's challenge.
        
        Returns:
            int: Shared key.
        """
        return self.key * challenge


from .blake2 import BLAKE2b, BLAKE2s
from .keccak import Keccak
from .md4 import MD4
from .md5 import MD5
from .ripemd160 import RIPEMD160
from .sha1 import SHA1
from .sha2 import SHA224, SHA256, SHA384, SHA512
from .sha3 import SHA3


__all__ = ["BLAKE2b", "BLAKE2s", "Keccak", "MD4", "MD5", "RIPEMD160", "SHA1", "SHA224", "SHA256", "SHA384", "SHA512", "SHA3"]


from samson.utilities.bytes import Bytes
from samson.hashes.sha1 import SHA1
from types import FunctionType

def MGF1(seed: bytes, length: int) -> Bytes:
    """
    Peforms the mask generation function v1 from RFC3447 B.2.1.

    Parameters:
        seed (bytes): Initial value.
        length (int): Length of mask to produce.
    
    Returns:
        Bytes: Mask.
    """
    mask = b''
    sha1 = SHA1()

    for i in range((length + 19) // 20):
        mask += sha1.hash(seed + Bytes(i).zfill(4))
    
    return mask[:length]


# https://www.ietf.org/rfc/rfc3447.txt
class OAEP(object):
    """
    Optimal Asymmetric Encryption Padding

    Probablistic Feistel Network proven to be semantically secure under chosen plaintext attack.
    """

    def __init__(self, modulus_len: int, mgf: FunctionType=MGF1, hash_obj: object=SHA1(), label: bytes=b''):
        """
        Parameters:
            modulus_len (int): Length of the RSA modulus, i.e. RSA bit strength.
            mgf        (func): Mask generation function. Takes in `seed` and `length` and returns bytes.
            hash_obj (object): Instantiated object with compatible hash interface.
            label     (bytes): (Optional) 
        """
        self.mgf = mgf
        self.hash_obj = hash_obj
        self.label = label
        self.modulus_len = modulus_len


    def __repr__(self):
        return f"<OAEP: mgf={self.mgf}, hash_obj={self.hash_obj}, label={self.label}>"

    def __str__(self):
        return self.__repr__()

        

    def pad(self, plaintext: bytes, seed: bytes=None) -> Bytes:
        """
        Pads the `plaintext`.

        Parameters:
            plaintext (bytes): Plaintext to pad.
            seed      (bytes): (Optional) Random seed for the MGF.
        
        Returns:
            Bytes: Padded plaintext.
        """
        plaintext = Bytes.wrap(plaintext)
        k = (self.modulus_len + 7) // 8

        # Step 1: Length checking
        h_len = self.hash_obj.digest_size
        m_len = len(plaintext)
        ps_len = k - m_len - (2 * h_len) - 2

        if ps_len < 0:
            raise ValueError("Plaintext is too long")
        

        # Step 2: EME-OAEP encoding
        l_hash = self.hash_obj.hash(self.label)

        ps = Bytes(b'').zfill(ps_len)
        db = l_hash + ps + b'\x01' + plaintext

        seed = seed or Bytes.random(h_len)

        db_mask = self.mgf(seed, k - h_len - 1)
        masked_db = db ^ db_mask

        seed_mask = self.mgf(masked_db, h_len)
        masked_seed = seed ^ seed_mask

        return b'\x00' + masked_seed + masked_db



    def unpad(self, plaintext: bytes, allow_mangers: bool=False, skip_label_check: bool=False) -> Bytes:
        """
        Unpads the `plaintext`.

        Parameters:
            plaintext         (bytes): Plaintext to pad.
            allow_mangers      (bool): Whether or not to explicitly help Manger's attack.
            skip_label_check   (bool): Whether or not to skip checking the label.
        
        Returns:
            Bytes: Unpadded plaintext.
        """
        k = (self.modulus_len + 7) // 8
        h_len = self.hash_obj.digest_size
        plaintext = Bytes.wrap(plaintext).zfill(k)

        if allow_mangers:
            if plaintext[0] != 0:
                raise ValueError("First byte is not zero! ;)")
        
        masked_seed, masked_db = plaintext[1:(h_len + 1)], plaintext[(h_len + 1):]

        seed_mask = self.mgf(masked_db, h_len)
        seed = masked_seed ^ seed_mask

        db_mask = self.mgf(seed, k - h_len - 1)
        db = masked_db ^ db_mask

        l_hash, m = db[:h_len], db[h_len + db[h_len:].index(b'\x01') + 1:]
        if not skip_label_check and l_hash != self.hash_obj.hash(self.label):
            raise ValueError("Label hashes do not match")

        return m


from samson.utilities.bytes import Bytes
from samson.publickey.dsa import DSA
from samson.utilities.ecc import EdwardsCurve25519, TwistedEdwardsPoint, TwistedEdwardsCurve
from samson.hashes.sha2 import SHA512
from copy import deepcopy

def bit(h,i):
  return (h[i//8] >> (i%8)) & 1

# https://ed25519.cr.yp.to/python/ed25519.py
class EdDSA(DSA):
    """
    Edwards Curve Digitial Signature Algorithm
    """

    def __init__(self, curve: TwistedEdwardsCurve=EdwardsCurve25519, hash_obj: object=SHA512(), d: int=None, A: TwistedEdwardsPoint=None):
        """
        Parameters:
            curve (TwistedEdwardsCurve): Curve used for calculations.
            hash_obj           (object): Instantiated object with compatible hash interface.
            d                     (int): Private key.
        """
        self.B = curve.B
        self.curve = curve
        self.d = Bytes.wrap(d or max(1, Bytes.random((curve.b + 7) // 8).int() % curve.q))
        self.H = hash_obj

        self.h = hash_obj.hash(self.d)

        self.a = 2**(curve.n) | sum(2**i * bit(self.h, i) for i in range(curve.c, curve.n))
        self.A = A or self.B * self.a



    def __repr__(self):
        return f"<EdDSA: d={self.d}, A={self.A}, curve={self.curve}, H={self.H}>"

    def __str__(self):
        return self.__repr__()
    

    def encode_point(self, P: TwistedEdwardsPoint) -> Bytes:
        """
        Encodes a `TwistedEdwardsPoint` as `Bytes`.

        Parameters:
            P (TwistedEdwardsPoint): Point to encode.
        
        Returns:
            Bytes: `Bytes` encoding.
        """
        x, y = P.x, P.y
        return Bytes(((x & 1) << self.curve.b-1) + ((y << 1) >> 1), 'little').zfill(self.curve.b // 8)
    


    def decode_point(self, in_bytes: Bytes) -> TwistedEdwardsPoint:
        """
        Decodes `Bytes` to a `TwistedEdwardsPoint`.

        Parameters:
            in_bytes (Bytes): `TwistedEdwardsPoint` encoded as `Bytes`.
        
        Returns:
            TwistedEdwardsPoint: Decoded point.
        """
        y_bytes = deepcopy(in_bytes)
        y_bytes[-1] &= 0x7F
        y = y_bytes.int()
        x = self.curve.recover_point_from_y(y).x

        if (x & 1) != bit(in_bytes, self.curve.b-1):
            x = self.curve.q - x
        
        return TwistedEdwardsPoint(x, y, self.curve)
        


    def sign(self, message: bytes) -> (int, int):
        """
        Signs a `message`.

        Parameters:
            message (bytes): Message to sign.
            k         (int): (Optional) Ephemeral key.
        
        Returns:
            (int, int): Signature formatted as (r, s).
        """
        r = self.H.hash(self.curve.magic + self.h[self.curve.b//8:] + message)[::-1].int()
        R = self.B * (r % self.curve.l)
        eR = self.encode_point(R)
        k = self.H.hash(self.curve.magic + eR + self.encode_point(self.A) + message)[::-1].int()
        S = (r + (k % self.curve.l) * self.a) % self.curve.l
        return eR + Bytes(S, 'little').zfill(self.curve.b//8)
    
    

    def verify(self, message: bytes, sig: (int, int)) -> bool:
        """
        Verifies a `message` against a `sig`.

        Parameters:
            message  (bytes): Message.
            sig ((int, int)): Signature of `message`.
        
        Returns:
            bool: Whether the signature is valid or not.
        """
        sig = Bytes.wrap(sig, 'little')

        if len(sig) != self.curve.b // 4:
            raise ValueError("`sig` length is wrong.")
        
        R = self.decode_point(sig[:self.curve.b//8])
        S = sig[self.curve.b//8:].int()

        h = self.H.hash(self.curve.magic + self.encode_point(R) + self.encode_point(self.A) + message)[::-1].int()

        return self.B * S == R + (self.A * h)


from fastecdsa.curve import P256
from fastecdsa.point import Point
from samson.utilities.general import rand_bytes

class ECDHE(object):
    """
    Elliptical curve Diffie-Hellman (Ephemeral).
    """

    def __init__(self, key: int=None, G: Point=P256.G):
        """
        Parameters:
            key (int): Secret key.
            G (Point): Generator point on an elliptical curve.
        """
        self.key = key or int.from_bytes(rand_bytes(), 'big')
        self.G = G



    def __repr__(self):
        return f"<ECDHE: key={self.key}, G={self.G}>"

    def __str__(self):
        return self.__repr__()



    def get_challenge(self) -> Point:
        """
        Gets the challenge.

        Returns:
            Point: The challenge.
        """
        return self.key * self.G



    def derive_key(self, challenge: Point) -> Point:
        """
        Derives the shared key from the other instance's challenge.

        Parameters:
            challenge (Point): The other instance's challenge.
        
        Returns:
            Point: Shared key.
        """
        return self.key * challenge


from samson.utilities.math import gcd, lcm, mod_inv, find_prime
from samson.utilities.encoding import pem_encode, pem_decode
from samson.utilities.bytes import Bytes
from pyasn1.codec.der import encoder, decoder
from pyasn1.type.univ import Sequence, Integer
import random

class RSA(object):
    """
    Rivest-Shamir-Adleman public key cryptosystem
    """

    def __init__(self, bits: int, p: int=None, q: int=None, e: int=None):
        """
        Parameters:
            bits (int): Number of bits for strength and capacity.
            p (int): Secret prime modulus.
            q (int): Secret prime modulus.
            e (int): Public expoonent.
        """
        if p and q and not e:
            raise Exception("Argument 'e' must be set if 'p' and 'q' are set. ")

        self.e = e or 3
        phi = 0

        if p and q:
            phi = lcm(p - 1, q - 1)
            self.n = p * q

            if gcd(self.e, phi) != 1:
                raise Exception("Invalid 'p' and 'q': GCD(e, phi) != 1")
        else:
            next_p = p
            next_q = q
            while gcd(self.e, phi) != 1:
                if not p:
                    next_p = find_prime(bits // 2)
                
                if not q:
                    next_q = find_prime(bits // 2)

                phi = lcm(next_p - 1, next_q - 1)

            p = next_p
            q = next_q
            self.n = p * q

        self.p = p
        self.q = q

        self.bits = bits

        self.phi = phi
        self.d = mod_inv(self.e, phi)
        self.alt_d = mod_inv(self.e, (self.p - 1) * (self.q - 1))

        self.pub = (self.e, self.n)
        self.priv = (self.d, self.n)



    def __repr__(self):
        return f"<RSA: bits={self.bits}, p={self.p}, q={self.q}, e={self.e}, n={self.n}, phi={self.phi}, d={self.d}, alt_d={self.alt_d}>"

    def __str__(self):
        return self.__repr__()


        
    def encrypt(self, plaintext: bytes) -> int:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Plaintext.
        
        Returns:
            int: Ciphertext.
        """
        m = Bytes.wrap(plaintext).int()
        return pow(m, self.e, self.n)



    def decrypt(self, ciphertext: int) -> Bytes:
        """
        Decrypts `ciphertext` back into plaintext.

        Parameters:
            ciphertext (int): Ciphertext.
        
        Returns:
            Bytes: Decrypted plaintext.
        """
        plaintext = pow(ciphertext, self.d, self.n)
        return Bytes(plaintext, 'big')



    def export_key(self, encode_pem: bool=True) -> bytes:
        """
        Exports the RSA instance into DER-encoded bytes.

        Parameters:
            encode_pem (bool): Whether or not to PEM-encode as well.
        
        Returns:
            bytes: DER-encoding of RSA instance.
        """
        seq = Sequence()

        for x in [0, self.n, self.e, self.d, self.p, self.q]:
            seq.setComponentByPosition(len(seq), Integer(x))
        
        der_encoded = encoder.encode(seq)

        if encode_pem:
            der_encoded = pem_encode(der_encoded, 'RSA PRIVATE KEY')

        return der_encoded


    
    @staticmethod
    def import_key(buffer: bytes):
        """
        Builds an RSA instance from DER and/or PEM-encoded bytes.

        Parameters:
            buffers (bytes): DER and/or PEM-encoded bytes.
        
        Returns:
            RSA: RSA instance.
        """
        try:
            buffer = pem_decode(buffer)
        except ValueError as _:
            pass
        

        seq = decoder.decode(buffer)
        _n, e, _d, p, q = [int(item) for item in seq[0][1:]]
        rsa = RSA(0, p=p, q=q, e=e)
        rsa.bits = rsa.n.bit_length()
        return rsa
    



    @staticmethod
    def factorize_from_shared_p(n1: int, n2: int, e: int):
        """
        Factorizes the moduli of two instances that share a common secret prime. See `Batch GCD`.

        Parameters:
            n1 (int): Modulus of the first instance.
            n2 (int): Modulus of the second instance.
            e  (int): Public exponent.
        
        Returns:
            (RSA, RSA): Both cracked RSA instances.
        """
        assert n1 != n2

        # Find shared `p`
        p = gcd(n1, n2)

        q1 = n1 // p
        q2 = n2 // p

        return (RSA(0, p=p, q=q1, e=e), RSA(0, p=p, q=q2, e=e))


    @staticmethod
    def factorize_from_faulty_crt(message: int, faulty_sig: int, e: int, n: int):
        """
        Factorize the secret primes from a faulty signature produced with CRT-optimized RSA.

        Parameters:
            message    (int): Message.
            faulty_sig (int): Faulty signature of `message`.
            e          (int): Public exponent.
            n          (int): Modulus.
        
        Returns:
            RSA: Cracked RSA instance.
        """
        q = gcd(pow(faulty_sig, e, n) - message, n)
        p = n // q

        return RSA(0, p=p, q=q, e=e)


    @staticmethod
    def factorize_from_d(d: int, e: int, n: int):
        """
        Factorizes the secret primes from the private key `d`.

        Parameters:
            d (int): Private key.
            e (int): Public exponent.
            n (int): Modulus.
        
        Returns:
            RSA: Full RSA instance.
        """
        k = d*e - 1
        p = None
        q = None

        while not p:
            g = random.randint(2, n - 1)
            t = k

            while t % 2 == 0:
                t = t // 2
                x = pow(g, t, n)

                if x > 1 and gcd(x - 1, n) > 1:
                    p = gcd(x - 1, n)
                    q = n // p
                    break
            
        return RSA(0, p=p, q=q, e=e)


from types import FunctionType

class LFG(object):
    """
    Lagged Fibonacci generator
    """

    ADD_OP = lambda a, b: a + b
    SUB_OP = lambda a, b: a - b

    def __init__(self, state: int, tap: int, feed: int, length: int, operation: FunctionType=ADD_OP):
        """
        Parameters:
            state     (int): Initial state.
            tap       (int): Initial tap position.
            feed      (int): Initial feed position.
            length    (int): Length of internal state (modulus).
            operation (int): The operation the LFG performs. Function that takes in an integer and returns an integer.
        """
        self.state = state
        self.tap = tap
        self.feed = feed
        self.length = length
        self.operation = operation


    def __repr__(self):
        return f"<LFG: state={self.state}, tap={self.tap}, feed={self.feed}, length={self.length}, operation={self.operation}>"

    def __str__(self):
        return self.__repr__()



    def generate(self) -> int:
        """
        Generates the next psuedorandom output.

        Returns:
            int: Next psuedorandom output.
        """
        self.tap = (self.tap - 1) % self.length
        self.feed = (self.feed - 1) % self.length

        x = self.operation(self.state[self.feed], self.state[self.tap]) & 0xFFFFFFFFFFFFFFFF
        self.state[self.feed] = x
        return x


from samson.utilities.bytes import Bytes
from samson.oracles.padding_oracle import PaddingOracle
from samson.publickey.rsa import RSA
import math

import logging
log = logging.getLogger(__name__)


# AKA OAEP padding oracle
# http://archiv.infsec.ethz.ch/education/fs08/secsem/Manger01.pdf
class MangersAttack(object):
    """
    Performs a plaintext recovery attack.
    
    Manger's attack stems from a padding oracle on RSA-OAEP. According to OAEP's specification, the first byte
    *has* to be zero. If the code checks this and leaks whether it's zero or not, we can efficiently retrieve
    the plaintext through an adaptive chosen-plaintext attack.
    """

    def __init__(self, padding_oracle: PaddingOracle, rsa: RSA):
        """
        Parameters:
            padding_oracle (PaddingOracle): An oracle that takes in bytes and returns whether the first byte of the decrypted plaintext is zero.
            rsa                      (RSA): An RSA instance containing the public key parameters.
        """
        self.oracle = padding_oracle
        self.rsa = rsa


    def _greater_equal_B(self, f, c, e, N):
        f_e = pow(f, e, N)
        return self.oracle.check_padding(Bytes((c * f_e) % N))


    def execute(self, ciphertext: bytes) -> Bytes:
        """
        Executes Manger's attack.

        Parameters:
            ciphertext (bytes): The ciphertext to decrypt.
        
        Returns:
            Bytes: The ciphertext's corresponding plaintext.
        """
        ciphertext = Bytes.wrap(ciphertext)
        ct_int = ciphertext.int()

        k = math.ceil(math.log(self.rsa.n, 256))
        B = 2 ** (8 * (k - 1))
        n = self.rsa.n
        e = self.rsa.e

        log.debug(f"k: {k}, B: {B}, n: {n}, e: {e}")
        
        # Step 1
        f1 = 2
        log.info("Starting step 1")

        while not self._greater_equal_B(f1, ct_int, e, n):
            f1 *= 2
        
        f1 //= 2
        log.debug(f"Found f1: {f1}")


        # Step 2
        nB = n + B
        nB_B = nB // B
        f2 = nB_B * f1

        log.info("Starting step 2")
        while self._greater_equal_B(f2, ct_int, e, n):
            f2 += f1
        

        log.debug(f"Found f2: {f2}")

        # Step 3
        div_mod = 1 if n % f2 else 0
        m_min = n // f2 + div_mod
        m_max = nB // f2
        BB = 2*B
        diff = m_max - m_min
        ctr = 0

        log.info("Starting step 3")
        log.debug(f"B-(diff * f2) = {B - (diff * f2)}")
        while diff > 0:
            if ctr % 100 == 0:
                log.debug(f"Iteration {ctr} difference: {diff}")
            
            f = BB // diff
            f_min = f * m_min
            i = f_min // n
            iN = i*n

            div_mod = 1 if iN % m_min else 0
            f3 = iN // m_min + div_mod
            iNB = iN + B

            if self._greater_equal_B(f3, ct_int, e, n):
                div_mod = 1 if iNB % f3 else 0
                m_min = iNB // f3 + div_mod
            else:
                m_max = iNB // f3
            
            diff = m_max - m_min
            ctr += 1


        return Bytes(m_min)


from math import ceil
from samson.utilities.bytes import Bytes
from types import FunctionType

class SpongeConstruction(object):
    """
    "The sponge construction is a mode of operation, based on a fixed-length permutation (or transformation) and on a padding rule, which builds a function mapping variable-length input to variable-length output"
    https://keccak.team/sponge_duplex.html
    """

    def __init__(self, perm_func: FunctionType, pad_func: FunctionType, r: int, c: int):
        """
        Parameters:
            perm_func (func): Sponge permutation function.
            pad_func  (func): Function for padding messages. Takes in bytes and returns them with padding.
            r          (int): Bit-size of the sponge function.
            c          (int): Sponge capacity.
        """
        self.perm_func = perm_func
        self.pad_func = pad_func

        self.r = r
        self.c = c

        self.block_size = (self.r + 7) // 8

        self.S = [[0] * 5 for _ in range(5)]


    def __repr__(self):
        return f"<Sponge r={self.r}, c={self.c}, block_size={self.block_size}, S={self.S}, pad_func={self.pad_func}>"

    def __str__(self):
        return self.__repr__()
    

    def reset(self):
        """
        Resets the sponge to its initial state.
        """
        self.S = [[0] * 5 for _ in range(5)]

    

    def absorb(self, in_bytes: bytes):
        """
        Absorbs bytes into the sponge.

        Parameters:
            in_bytes (bytes): Bytes to absorb.
        """
        padded = self.pad_func(in_bytes)
        
        for block in padded.chunk(self.block_size):
            curr = 0
            for y in range(5):
                for x in range(5):
                    self.S[x][y] ^= sum([byte << i * 8 for i, byte in enumerate(block[curr:curr + 8])])
                    curr += 8

            self.S = self.perm_func(self.S)

    

    def squeeze(self, amount: int):
        """
        Squeezes bytes out of the sponge.

        Parameters:
            amount (int): Number of bytes to return.
        
        Return:
            generator: Yields blocks of Bytes until at least request `amount` has been returned.
        """
        for _ in range(ceil(amount / self.block_size)):
            out = Bytes(b'')
            for y in range(5):
                for x in range(5):
                    out += Bytes(self.S[x][y]).zfill(8)[::-1]

            yield out[:self.block_size]
            self.S = self.perm_func(self.S)



from samson.utilities.math import mod_inv, lll, generate_superincreasing_seq, find_coprime
from samson.utilities.bytes import Bytes
from sympy.matrices import Matrix, eye


class MerkleHellmanKnapsack(object):
    """
    Merkle-Hellman Knapsack cryptosystem.

    Broken and extremely rare cryptosystem based off of the hardness of the Knapsack problem.
    """

    def __init__(self, priv: list=None, q: int=None, r: int=None, max_diff: int=2**20, key_len: int=8):
        """
        Parameters:
            priv    (list): Private key made from a superincreasing sequence.
            q        (int): Modulus. Integer greater than the sum of `priv`.
            r        (int): Multiplier. Integer coprime to `q` and between `q // 4` and `q`.
            max_diff (int): Maximum difference between integers in the superincreasing sequence. Used for generating `priv`.
            key_len  (int): Desired length of the key.
        """
        super_seq = generate_superincreasing_seq(key_len + 1, max_diff, starting=Bytes.random(max(1, key_len // 8)).int())
        self.priv = priv or super_seq[:key_len]
        self.q = q or super_seq[-1]
        self.r = r or find_coprime(self.q, range(self.q // 4, self.q))

        self.pub = [(w * self.r) % self.q for w in self.priv]



    def __repr__(self):
        return f"<MerkleHellmanKnapsack: priv={self.priv}, pub={self.pub}, q={self.q}, r={self.r}>"
    
    def __str__(self):
        return self.__repr__()



    def encrypt(self, message: bytes) -> list:
        """
        Encrypt `message`.

        Parameters:
            message (bytes): Message to encrypt.

        Returns:
            list: List of ciphertext integers. List cardinality dependent on message/key-length ratio.
        """
        bin_str = ''
        for byte in message:
            bin_str += bin(byte)[2:].zfill(8)

        all_sums = []
        
        for i in range(len(bin_str) // len(self.pub)):
            byte_str = bin_str[i * len(self.pub):(i + 1) * len(self.pub)]
            all_sums.append(sum([int(byte_str[j]) * self.pub[j] for j in range(len(self.pub))]))

        return all_sums



    def decrypt(self, sums: list) -> Bytes:
        """
        Decrypts `sums` back into plaintext.

        Parameters:
            sums (list): List of ciphertext sums.
        
        Returns:
            Bytes: Decrypted plaintext.
        """
        r_inv = mod_inv(self.r, self.q)
        inv_sums = [(byte_sum * r_inv) % self.q for byte_sum in sums]
        plaintext = Bytes(b'')

        for inv_sum in inv_sums:
            curr = inv_sum
            bin_string = ''

            for i in range(len(self.pub) - 1, -1, -1):
                if self.priv[i] <= curr:
                    curr -= self.priv[i]
                    bin_string += '1'
                else:
                    bin_string += '0'

            plaintext += int.to_bytes(int(bin_string[::-1], 2), len(self.pub) // 8, 'big')

        return plaintext


    @staticmethod
    def recover_plaintext(ciphertext: int, pub: list, alpha: int=1) -> Bytes:
        """
        Attempts to recover the plaintext without the private key.

        Parameters:
            ciphertext (int): A ciphertext sum.
            pub        (int): The public key.
            alpha      (int): Punishment coefficient for deviation from guessed bit distribution.
        
        Returns:
            Bytes: Recovered plaintext.
        """
        ident = eye(len(pub))
        pub_matrix = ident.col_join(Matrix([pub]))
        problem_matrix = pub_matrix.row_join(Matrix([[0] * len(pub) + [-ciphertext]]).T)

        for i in range(len(pub)):
            alpha_penalizer = Matrix([alpha] * len(pub) + [-alpha * i]).T
            problem_matrix_prime = problem_matrix.col_join(alpha_penalizer).T
            matrices = [problem_matrix_prime.row(row) for row in range(problem_matrix_prime.rows)]

            solution_matrix = lll(matrices, 0.99)

            for row in range(solution_matrix.rows):
                row_mat = solution_matrix.row(row)
                new_row = [item for item in row_mat if item >= 0 and item <= 1]

                if len(new_row) == len(row_mat):
                    return Bytes(int(''.join([str(val) for val in row_mat[:-2]]), 2))
        return solution_matrix


from samson.utilities.bytes import Bytes
from types import FunctionType

class PBKDF1(object):
    """
    Password-based Key Derivation Function v1.
    """

    def __init__(self, hash_fn: FunctionType, desired_len: int, num_iters: int):
        """
        Parameters:
            hash_fn    (func): Function that takes in bytes and returns them hashed.
            desired_len (int): Desired output length.
            num_iters   (int): Number of iterations to perform.
        """
        self.hash_fn = hash_fn
        self.num_iters = num_iters
        self.desired_len = desired_len


    def __repr__(self):
        return f"<PBKDF1: hash_fn={self.hash_fn}, desired_len={self.desired_len} num_iters={self.num_iters}>"

    def __str__(self):
        return self.__repr__()


    def derive(self, password: bytes, salt: bytes) -> Bytes:
        """
        Derives a key.

        Parameters:
            password (bytes): Bytes-like object to key the internal state.
            salt     (bytes): Salt to tweak the output.
        
        Returns:
            Bytes: Derived key.
        """
        last_result = password + salt
        for _ in range(self.num_iters):
            last_result = self.hash_fn(last_result)
        
        return last_result[:self.desired_len]


import string

# http://practicalcryptography.com/ciphers/playfair-cipher/
class Playfair(object):
    """
    Bigram substituion cipher.
    """

    def __init__(self, key: str):
        """
        Parameters:
            key (str): Keyphrase containing no duplicate letters and not the letter 'j'.
        """
        if len(list(set(list(key)))) < len(key):
            raise Exception("Key cannot have duplicate characters")

        # Playfair combines 'i' and 'j'
        assert 'j' not in key

        for letter in string.ascii_lowercase:
            if letter not in key and letter != 'j':
                key += letter

        self.key = key


    def __repr__(self):
        return "<Playfair key={}>".format(self.key)

    def __str__(self):
        return self.__repr__()



    def encrypt(self, plaintext: str) -> str:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (str): String to be encrypted.
        
        Returns:
            str: Resulting ciphertext.
        """
        plaintext = plaintext.replace("j", "i")

        # Remove characters not in key
        for char in plaintext:
            if char not in self.key:
                plaintext = plaintext.replace(char, "")


        # If plaintext has an odd number of characters, add an 'x' to the end
        if len(plaintext) % 2 == 1:
            plaintext += 'x'


        # Substitute duplicate letters with x
        for i in range(0, len(plaintext), 2):
            if plaintext[i] == plaintext[i + 1]:
                plaintext = plaintext[:i + 1] + 'x' + plaintext[i + 2:]
        

        ciphertext = ''
        for i in range(0, len(plaintext), 2):
            a, b = plaintext[i], plaintext[i + 1]
            a_loc, b_loc = self.key.index(a), self.key.index(b)

            a_row, a_col = a_loc // 5, a_loc % 5
            b_row, b_col = b_loc // 5, b_loc % 5

            if a_row != b_row and a_col != b_col:
                new_a = a_row * 5 + b_col
                new_b = b_row * 5 + a_col
            elif a_row == b_row:
                new_a = a_row * 5 + ((a_col + 1) % 5)
                new_b = a_row * 5 + ((b_col + 1) % 5)
            else:
                new_a = ((a_row + 1) % 5) * 5 + a_col
                new_b = ((b_row + 1) % 5) * 5 + a_col

            ciphertext += self.key[new_a] + self.key[new_b]
        return ciphertext



    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (str): Stringt to be decrypted.
        
        Returns:
            str: Resulting plaintext.
        """
        plaintext = ''
        for i in range(0, len(ciphertext), 2):
            a, b = ciphertext[i], ciphertext[i + 1]
            a_loc, b_loc = self.key.index(a), self.key.index(b)

            a_row, a_col = a_loc // 5, a_loc % 5
            b_row, b_col = b_loc // 5, b_loc % 5

            if a_row != b_row and a_col != b_col:
                new_a = a_row * 5 + b_col
                new_b = b_row * 5 + a_col
            elif a_row == b_row:
                new_a = a_row * 5 + ((a_col - 1) % 5)
                new_b = a_row * 5 + ((b_col - 1) % 5)
            else:
                new_a = ((a_row - 1) % 5) * 5 + a_col
                new_b = ((b_row - 1) % 5) * 5 + a_col

            plaintext += self.key[new_a] + self.key[new_b]
        return plaintext


from samson.block_ciphers.des import DES
from samson.utilities.bytes import Bytes

class TDES(object):
    """
    3DES in EDE mode.

    Structure: Feistel Network
    Key size: 64, 128, 192 bits (56, 112, 168 bits of security)
    Block size: 64 bits
    """

    def __init__(self, key: bytes):
        """
        Parameters:
            key (bytes): Bytes-like object to key the cipher.
        """
        key = Bytes.wrap(key)
        if not len(key) in [8, 16, 24]:
            raise ValueError('`key` size must be in [8, 16, 24]')

        self.key = key
        self.des_arr = [DES(subkey.zfill(8)) for subkey in key.chunk(8)]
        self.block_size = 8
        

    
    def __repr__(self):
        return f"<TDES: key={self.key}, des_arr={self.des_arr}>"

    def __str__(self):
        return self.__repr__()

    
    def encrypt(self, plaintext: bytes) -> Bytes:
        """
        Encrypts `plaintext`.

        Parameters:
            plaintext (bytes): Bytes-like object to be encrypted.
        
        Returns:
            Bytes: Resulting ciphertext.
        """
        plaintext = Bytes.wrap(plaintext)
        pt_1 = self.des_arr[0].encrypt(plaintext)
        pt_2 = self.des_arr[1].decrypt(pt_1)
        ciphertext = self.des_arr[2].encrypt(pt_2)

        return ciphertext


    def decrypt(self, ciphertext: bytes) -> Bytes:
        """
        Decrypts `ciphertext`.

        Parameters:
            ciphertext (bytes): Bytes-like object to be decrypted.
        
        Returns:
            Bytes: Resulting plaintext.
        """
        ciphertext = Bytes.wrap(ciphertext)
        ct_1 = self.des_arr[2].decrypt(ciphertext)
        ct_2 = self.des_arr[1].encrypt(ct_1)
        plaintext = self.des_arr[0].decrypt(ct_2)

        return plaintext


from types import FunctionType

class DefaultOracle(object):
    """
    Simple oracle that provides a `request` function.
    """

    def __init__(self, request_func: FunctionType):
        """
        Parameters:
            request_func (func): Function that provides the oracle.
        """
        self.request = request_func